"""
pyart.retrieve.spectra
======================

Retrievals from spectral data.

.. autosummary::
    :toctree: generated/

    compute_iq
    compute_spectral_power
    compute_spectral_noise
    compute_spectral_reflectivity
    compute_spectral_differential_reflectivity
    compute_spectral_differential_phase
    compute_spectral_rhohv
    compute_spectral_phase
    compute_pol_variables
    compute_noise_power
    compute_reflectivity
    compute_differential_reflectivity
    compute_differential_phase
    compute_rhohv
    compute_Doppler_velocity
    compute_Doppler_width
    _compute_power
    _smooth_spectral_power

"""
from copy import deepcopy
from warnings import warn

import numpy as np
from scipy.signal.windows import gaussian
from scipy.signal.windows import get_window

from ..config import get_metadata, get_field_name
from ..util import radar_from_spectra, rolling_window, estimate_noise_hs74
from ..util import ma_broadcast_to

def compute_iq(spectra, fields_in_list, fields_out_list, window=None):
    """
    Computes the IQ data from the spectra through an inverse Fourier transform

    Parameters
    ----------
    spectra : Spectra radar object
        Object containing the spectra
    fields_in_list : list of str
        list of input spectra fields names
    fields_out_list : list of str
        list with the output IQ fields names obtained from the input fields
    window : string, tupple or None
        Parameters of the window used to obtain the spectra. The parameters
        are the ones corresponding to function
        scipy.signal.windows.get_window. If it is not None the inverse will be
        used to multiply the IQ data obtained by the IFFT

    Returns
    -------
    radar : IQ radar object
        radar object containing the IQ fields

    """
    radar = deepcopy(spectra)
    radar.fields = {}
    for field_name_in, field_name_out in zip(fields_in_list, fields_out_list):
        if field_name_out in ('IQ_hh_ADU', 'IQ_vv_ADU'):
            iq = np.ma.masked_all(
                (spectra.nrays, spectra.ngates, spectra.npulses_max),
                dtype=np.complex64)
            for ray, npuls in enumerate(spectra.npulses['data']):
                ray_data = spectra.fields[field_name_in]['data'][
                    ray, :, 0:npuls].filled(0.)
                iq[ray, :, 0:npuls] = np.fft.ifft(np.fft.ifftshift(
                    ray_data, axes=-1), axis=-1)*npuls

                if window is not None:
                    wind = get_window(window, npuls)
                    wind = wind/np.sqrt(np.sum(np.power(wind, 2.))/npuls)
                    wind = np.broadcast_to(
                        np.atleast_2d(wind), (spectra.ngates, npuls))
                    iq[ray, :, 0:npuls] /= wind
        else:
            iq = np.ma.masked_all(
                (spectra.nrays, spectra.ngates, spectra.npulses_max),
                dtype=np.float32)
            for ray, npuls in enumerate(spectra.npulses['data']):
                iq[ray, :, 0:npuls] = spectra.fields[field_name_in]['data'][
                    ray, :, 0:npuls]*npuls

        field_dict = get_metadata(field_name_out)
        field_dict['data'] = iq
        radar.fields.update({field_name_out: field_dict})

    return radar


def compute_spectral_power(spectra, units='dBADU', subtract_noise=False,
                           smooth_window=None, signal_field=None,
                           noise_field=None):
    """
    Computes the spectral power from the complex spectra in ADU. Requires
    key dBADU_to_dBm_hh or dBADU_to_dBm_vv in radar_calibration if the
    units are to be dBm

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    units : str
        The units of the returned signal. Can be 'ADU', 'dBADU' or 'dBm'
    subtract_noise : Bool
        If True noise will be subtracted from the signal
    smooth_window : int or None
        Size of the moving Gaussian smoothing window. If none no smoothing
        will be applied
    signal_field, noise_field : str, optional
        Name of the fields in radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    pwr_dict : field dictionary
        Field dictionary containing the spectral power

    """
    if signal_field is None:
        signal_field = get_field_name('complex_spectra_hh_ADU')
    if noise_field is None:
        noise_field = get_field_name('spectral_noise_power_hh_ADU')

    pol = 'hh'
    if 'vv' in signal_field:
        pol = 'vv'

    noise = None
    if noise_field in spectra.fields:
        noise = spectra.fields[noise_field]['data']

    pwr = _compute_power(
        spectra.fields[signal_field]['data'], noise=noise,
        subtract_noise=subtract_noise, smooth_window=smooth_window)

    if units in ('dBADU', 'dBm'):
        pwr = 10.*np.ma.log10(pwr)

        if units == 'dBm':
            dBADU2dBm = None
            if spectra.radar_calibration is not None:
                if (pol == 'hh' and
                        'dBADU_to_dBm_hh' in spectra.radar_calibration):
                    dBADU2dBm = (
                        spectra.radar_calibration['dBADU_to_dBm_hh']['data'][
                            0])
                elif (pol == 'vv' and
                      'dBADU_to_dBm_vv' in spectra.radar_calibration):
                    dBADU2dBm = (
                        spectra.radar_calibration['dBADU_to_dBm_vv']['data'][0])

            if dBADU2dBm is None:
                raise ValueError(
                    'Unable to compute spectral power in dBm. ' +
                    'dBADU to dBm conversion factor unknown')

            # should it be divided by the number of pulses?
            pwr += dBADU2dBm

    power_field = 'spectral_power_'+pol+'_'+units
    if 'unfiltered' in signal_field:
        power_field = 'unfiltered_'+power_field

    pwr_dict = get_metadata(power_field)
    pwr_dict['data'] = pwr

    return pwr_dict


def compute_spectral_noise(spectra, units='dBADU', navg=1, rmin=0.,
                           nnoise_min=1, signal_field=None):
    """
    Computes the spectral noise power from the complex spectra in ADU.
    Requires key dBADU_to_dBm_hh or dBADU_to_dBm_vv in radar_calibration if
    the units are to be dBm. The noise is computed using the method described
    in Hildebrand and Sehkon, 1974.

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    units : str
        The units of the returned signal. Can be 'ADU', 'dBADU' or 'dBm'
    navg : int
        Number of spectra averaged
    rmin : int
        Range from which the data is used to estimate the noise
    nnoise_min : int
        Minimum number of samples to consider the estimated noise power valid
    signal_field : str, optional
        Name of the field in radar which contains the signal.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    noise_dict : field dictionary
        Field dictionary containing the spectral noise power

    References
    ----------
    P. H. Hildebrand and R. S. Sekhon, Objective Determination of the Noise
    Level in Doppler Spectra. Journal of Applied Meteorology, 1974, 13,
    808-811.

    """
    if signal_field is None:
        signal_field = get_field_name('complex_spectra_hh_ADU')

    pol = 'hh'
    if 'vv' in signal_field:
        pol = 'vv'

    ind_rmin = np.where(spectra.range['data'] >= rmin)[0]
    if ind_rmin.size == 0:
        warn('Unable to compute spectral noise. ' +
             'Range at which start gathering data '+str(rmin) +
             'km. larger than radar range')
        return None

    ind_rmin = ind_rmin[0]

    pwr = _compute_power(spectra.fields[signal_field]['data'])

    noise = np.ma.masked_all(
        (spectra.nrays, spectra.ngates, spectra.npulses_max))
    for ray, npuls in enumerate(spectra.npulses['data']):
        mean, _, _, _ = estimate_noise_hs74(
            pwr[ray, ind_rmin:, 0:npuls].compressed(), navg=navg,
            nnoise_min=nnoise_min)
        noise[ray, :, :] = mean
        print(mean)

    if units in ('dBADU', 'dBm'):
        noise = 10.*np.ma.log10(noise)

        if units == 'dBm':
            dBADU2dBm = None
            if spectra.radar_calibration is not None:
                if (pol == 'hh' and
                        'dBADU_to_dBm_hh' in spectra.radar_calibration):
                    dBADU2dBm = (
                        spectra.radar_calibration['dBADU_to_dBm_hh']['data'][
                            0])
                elif (pol == 'vv' and
                      'dBADU_to_dBm_vv' in spectra.radar_calibration):
                    dBADU2dBm = (
                        spectra.radar_calibration['dBADU_to_dBm_vv']['data'][0])

            if dBADU2dBm is None:
                raise ValueError(
                    'Unable to compute spectral power in dBm. ' +
                    'dBADU to dBm conversion factor unknown')

            # should it be divided by the number of pulses?
            noise += dBADU2dBm

    noise_field = 'spectral_noise_power_'+pol+'_'+units
    noise_dict = get_metadata(noise_field)
    noise_dict['data'] = noise

    return noise_dict


def compute_spectral_reflectivity(spectra, compute_power=True,
                                  subtract_noise=False, smooth_window=None,
                                  pwr_field=None, signal_field=None,
                                  noise_field=None):
    """
    Computes the spectral reflectivity from the complex spectra in ADU or
    from the signal power in ADU. Requires
    keys dBADU_to_dBm_hh or dBADU_to_dBm_vv in radar_calibration if the
    to be computed

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    compute_power : Bool
        If True the signal power will be computed. Otherwise the field given
        by the user will be used
    subtract_noise : Bool
        If True noise will be subtracted from the signal
    smooth_window : int or None
        Size of the moving Gaussian smoothing window. If none no smoothing
        will be applied
    pwr_field, signal_field, noise_field : str, optional
        Name of the fields in radar which contains the signal power, complex
        signal and noise. None will use the default field name in the Py-ART
        configuration file.

    Returns
    -------
    sdBZ_dict : field dictionary
        Field dictionary containing the spectral reflectivity

    """
    if compute_power:
        if signal_field is None:
            signal_field = get_field_name('complex_spectra_hh_ADU')
        if noise_field is None:
            noise_field = get_field_name('spectral_noise_power_hh_ADU')
    else:
        if pwr_field is None:
            pwr_field = get_field_name('spectral_power_hh_ADU')

    if spectra.radar_calibration is None:
        raise ValueError(
            'Unable to compute spectral reflectivity. ' +
            'Calibration parameters unknown')

    pol = 'hh'
    if ((signal_field is not None and 'vv' in signal_field) or
            (pwr_field is not None and 'vv' in pwr_field)):
        pol = 'vv'

    if pol == 'hh':
        if ('dBADU_to_dBm_hh' not in spectra.radar_calibration or
                'calibration_constant_hh' not in spectra.radar_calibration):
            raise ValueError(
                'Unable to compute spectral reflectivity. ' +
                'Calibration parameters unknown')
        dBADU2dBm = spectra.radar_calibration['dBADU_to_dBm_hh']['data'][0]
        radconst = (
            spectra.radar_calibration['calibration_constant_hh']['data'][0])
    else:
        if ('dBADU_to_dBm_vv' not in spectra.radar_calibration or
                'calibration_constant_vv' not in spectra.radar_calibration):
            raise ValueError(
                'Unable to compute spectral reflectivity. ' +
                'Calibration parameters unknown')
        dBADU2dBm = spectra.radar_calibration['dBADU_to_dBm_vv']['data'][0]
        radconst = (
            spectra.radar_calibration['calibration_constant_vv']['data'][0])

    if (pol == 'hh' and 'matched_filter_loss_h' in spectra.radar_calibration):
        mfloss = spectra.radar_calibration['matched_filter_loss_h']['data'][0]
    elif (pol == 'vv' and
          'matched_filter_loss_v' in spectra.radar_calibration):
        mfloss = spectra.radar_calibration['matched_filter_loss_v']['data'][0]
    else:
        warn('Unknown matched filter losses. Assumed 0 dB')
        mfloss = 0.

    if 'path_attenuation' in spectra.radar_calibration:
        pathatt = spectra.radar_calibration['path_attenuation']['data'][0]
    else:
        warn('Unknown gas path attenuation. Assumed 0 dB/km')
        pathatt = 0.

    rangeKm = np.broadcast_to(
        np.atleast_3d(spectra.range['data']/1000.),
        (spectra.nrays, spectra.ngates, spectra.npulses_max))

    if compute_power:
        noise = None
        if noise_field in spectra.fields:
            noise = spectra.fields[noise_field]['data']

        pwr = _compute_power(
            spectra.fields[signal_field]['data'], noise=noise,
            subtract_noise=subtract_noise, smooth_window=smooth_window)
    else:
        pwr = spectra.fields[pwr_field]['data']

    sdBZ = (
        10.*np.ma.log10(pwr)+dBADU2dBm+radconst+mfloss+pathatt*rangeKm +
        20.*np.log10(rangeKm))

    sdBZ_field = 'spectral_reflectivity_'+pol
    if ((signal_field is not None and 'unfiltered' in signal_field) or
            (pwr_field is not None and 'unfiltered' in pwr_field)):
        sdBZ_field = 'unfiltered_'+sdBZ_field

    sdBZ_dict = get_metadata(sdBZ_field)
    sdBZ_dict['data'] = sdBZ

    return sdBZ_dict


def compute_spectral_differential_reflectivity(spectra, compute_power=True,
                                               subtract_noise=False,
                                               smooth_window=None,
                                               pwr_h_field=None,
                                               pwr_v_field=None,
                                               signal_h_field=None,
                                               signal_v_field=None,
                                               noise_h_field=None,
                                               noise_v_field=None):
    """
    Computes the spectral differential reflectivity from the complex spectras
    or the power in ADU

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    compute_power : Bool
        If True the signal power will be computed. Otherwise the field given
        by the user will be used
    subtract_noise : Bool
        If True noise will be subtracted from the signals
    smooth_window : int or None
        Size of the moving Gaussian smoothing window. If none no smoothing
        will be applied
    pwr_h_field, pwr_v_field, signal_h_field, signal_v_field, noise_h_field,
    noise_v_field : str
        Name of the fields in radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    sZDR_dict : field dictionary
        Field dictionary containing the spectral differential reflectivity

    """
    if compute_power:
        if signal_h_field is None:
            signal_h_field = get_field_name('complex_spectra_hh_ADU')
        if signal_v_field is None:
            signal_v_field = get_field_name('complex_spectra_vv_ADU')
        if noise_h_field is None:
            noise_h_field = get_field_name('spectral_noise_power_hh_ADU')
        if noise_v_field is None:
            noise_v_field = get_field_name('spectral_noise_power_vv_ADU')
    else:
        if pwr_h_field is None:
            pwr_h_field = get_field_name('spectral_power_hh_ADU')
        if pwr_v_field is None:
            pwr_v_field = get_field_name('spectral_power_hh_ADU')

    if spectra.radar_calibration is None:
        raise ValueError(
            'Unable to compute spectral reflectivity. ' +
            'Calibration parameters unknown')

    if ('dBADU_to_dBm_hh' not in spectra.radar_calibration or
            'dBADU_to_dBm_vv' not in spectra.radar_calibration or
            'calibration_constant_hh' not in spectra.radar_calibration or
            'calibration_constant_vv' not in spectra.radar_calibration):
        raise ValueError(
            'Unable to compute spectral reflectivity. ' +
            'Calibration parameters unknown')

    dBADU2dBm_h = spectra.radar_calibration['dBADU_to_dBm_hh']['data'][0]
    dBADU2dBm_v = spectra.radar_calibration['dBADU_to_dBm_vv']['data'][0]
    radconst_h = (
        spectra.radar_calibration['calibration_constant_hh']['data'][0])
    radconst_v = (
        spectra.radar_calibration['calibration_constant_vv']['data'][0])

    if compute_power:
        noise = None
        if noise_h_field in spectra.fields:
            noise = spectra.fields[noise_h_field]['data']

        pwr_h = _compute_power(
            spectra.fields[signal_h_field]['data'], noise=noise,
            subtract_noise=subtract_noise, smooth_window=smooth_window)

        noise = None
        if noise_v_field in spectra.fields:
            noise = spectra.fields[noise_v_field]['data']

        pwr_v = _compute_power(
            spectra.fields[signal_v_field]['data'], noise=noise,
            subtract_noise=subtract_noise, smooth_window=smooth_window)
    else:
        pwr_h = spectra.fields[pwr_h_field]['data']
        pwr_v = spectra.fields[pwr_v_field]['data']

    sZDR = (
        (10.*np.ma.log10(pwr_h)+dBADU2dBm_h+radconst_h) -
        (10.*np.ma.log10(pwr_v)+dBADU2dBm_v+radconst_v))

    sZDR_field = 'spectral_differential_reflectivity'
    if ((signal_h_field is not None and 'unfiltered' in signal_h_field) or
            (pwr_h_field is not None and 'unfiltered' in pwr_h_field)):
        sZDR_field = 'unfiltered_'+sZDR_field

    sZDR_dict = get_metadata(sZDR_field)
    sZDR_dict['data'] = sZDR

    return sZDR_dict


def compute_spectral_differential_phase(spectra, use_rhohv=False,
                                        srhohv_field=None,
                                        signal_h_field=None,
                                        signal_v_field=None):
    """
    Computes the spectral differential reflectivity from the complex spectras
    in ADU or sRhoHV

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    use_rhohv : Bool
        If true sRhoHV is going to be used to compute the differential phase.
        Otherwise the complex signals are used
    signal_h_field, signal_v_field : str
        Name of the fields in radar which contains the signal.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    sPhiDP_dict : field dictionary
        Field dictionary containing the spectral differential phase

    """
    if use_rhohv:
        if srhohv_field is None:
            srhohv_field = get_field_name(
                'spectral_copolar_correlation_coefficient')
    else:
        if signal_h_field is None:
            signal_h_field = get_field_name('complex_spectra_hh_ADU')
        if signal_v_field is None:
            signal_v_field = get_field_name('complex_spectra_vv_ADU')

    if not use_rhohv:
        phase_h = np.ma.angle(
            spectra.fields[signal_h_field]['data'], deg=True)
        phase_v = np.ma.angle(
            spectra.fields[signal_v_field]['data'], deg=True)

        sPhiDP = phase_h-phase_v
    else:
        sPhiDP = np.ma.angle(
            spectra.fields[srhohv_field]['data'], deg=True)

    sPhiDP_field = 'spectral_differential_phase'
    if ((signal_h_field is not None and 'unfiltered' in signal_h_field) or
            (srhohv_field is not None and 'unfiltered' in srhohv_field)):
        sPhiDP_field = 'unfiltered_'+sPhiDP_field

    sPhiDP_dict = get_metadata(sPhiDP_field)
    sPhiDP_dict['data'] = sPhiDP

    return sPhiDP_dict


def compute_spectral_rhohv(spectra, subtract_noise=False, signal_h_field=None,
                           signal_v_field=None, noise_h_field=None,
                           noise_v_field=None):
    """
    Computes the spectral RhoHV from the complex spectras in ADU

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    subtract_noise : Bool
        If True noise will be subtracted from the signals
    signal_h_field, signal_v_field, noise_h_field, noise_v_field : str
        Name of the fields in radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    sRhoHV_dict : field dictionary
        Field dictionary containing the spectral RhoHV

    """
    if signal_h_field is None:
        signal_h_field = get_field_name('complex_spectra_hh_ADU')
    if signal_v_field is None:
        signal_v_field = get_field_name('complex_spectra_vv_ADU')
    if noise_h_field is None:
        noise_h_field = get_field_name('spectral_noise_power_hh_ADU')
    if noise_v_field is None:
        noise_v_field = get_field_name('spectral_noise_power_vv_ADU')

    sRhoHV = (
        spectra.fields[signal_h_field]['data'] *
        np.ma.conjugate(spectra.fields[signal_v_field]['data']))

    noise = None
    if noise_h_field in spectra.fields:
        noise = spectra.fields[noise_h_field]['data']

    pwr_h = _compute_power(
        spectra.fields[signal_h_field]['data'], noise=noise,
        subtract_noise=subtract_noise)

    noise = None
    if noise_v_field in spectra.fields:
        noise = spectra.fields[noise_v_field]['data']

    pwr_v = _compute_power(
        spectra.fields[signal_v_field]['data'], noise=noise,
        subtract_noise=subtract_noise)

    sRhoHV /= np.ma.sqrt(pwr_h*pwr_v)

    sRhoHV_field = 'spectral_copolar_correlation_coefficient'
    if 'unfiltered' in signal_h_field:
        sRhoHV_field = 'unfiltered_'+sRhoHV_field

    sRhoHV_dict = get_metadata(sRhoHV_field)
    sRhoHV_dict['data'] = sRhoHV

    return sRhoHV_dict


def compute_spectral_phase(spectra, signal_field=None):
    """
    Computes the spectral phase from the complex spectra in ADU

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    signal_field : str, optional
        Name of the field in radar which contains the signal.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    phase_dict : field dictionary
        Field dictionary containing the spectral phase

    """
    if signal_field is None:
        signal_field = get_field_name('complex_spectra_hh_ADU')

    pol = 'hh'
    if 'vv' in signal_field:
        pol = 'vv'

    phase = np.ma.angle(spectra.fields[signal_field]['data'], deg=True)

    phase_field = 'spectral_phase_'+pol
    if 'unfiltered' in signal_field:
        phase_field = 'unfiltered_'+phase_field

    phase_dict = get_metadata(phase_field)
    phase_dict['data'] = phase

    return phase_dict


def compute_pol_variables(spectra, fields_list, use_pwr=False,
                          subtract_noise=False, smooth_window=None,
                          srhohv_field=None, pwr_h_field=None,
                          pwr_v_field=None, signal_h_field=None,
                          signal_v_field=None, noise_h_field=None,
                          noise_v_field=None):
    """
    Computes the polarimetric variables from the complex spectra in ADU or
    the spectral powers and spectral RhoHV

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    fields_list : list of str
        list of fields to compute
    use_pwr : Bool
        If True the polarimetric variables will be computed from the spectral
        power and the spectral RhoHV. Otherwise from the complex spectra
    subtract_noise : Bool
        If True noise will be subtracted from the signals
    smooth_window : int or None
        Size of the moving Gaussian smoothing window. If none no smoothing
        will be applied
    srhohv_field, pwr_h_field, pwr_v_field, signal_h_field, signal_v_field,
    noise_h_field, noise_v_field : str
        Name of the fields in radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    radar : radar object
        Object containing the computed fields

    """
    if use_pwr:
        if srhohv_field is None:
            srhohv_field = get_field_name(
                'spectral_copolar_correlation_coefficient')
        if pwr_h_field is None:
            pwr_h_field = get_field_name('spectral_power_hh_ADU')
        if pwr_v_field is None:
            pwr_v_field = get_field_name('spectral_power_vv_ADU')

        compute_power = False
        use_rhohv = True
    else:
        if signal_h_field is None:
            signal_h_field = get_field_name('complex_spectra_hh_ADU')
        if signal_v_field is None:
            signal_v_field = get_field_name('complex_spectra_vv_ADU')
        if noise_h_field is None:
            noise_h_field = get_field_name('spectral_noise_power_hh_ADU')
        if noise_v_field is None:
            noise_v_field = get_field_name('spectral_noise_power_vv_ADU')

        compute_power = True
        use_rhohv = False

    fields = {}
    if ('reflectivity' in fields_list or
            'differential_reflectivity' in fields_list or
            'uncorrected_differential_phase' in fields_list or
            'velocity' in fields_list or 'spectrum_width' in fields_list or
            'unfiltered_reflectivity' in fields_list or
            'unfiltered_differential_reflectivity' in fields_list or
            'uncorrected_unfiltered_differential_phase' in fields_list or
            'unfiltered_velocity' in fields_list or
            'unfiltered_spectrum_width' in fields_list):
        sdBZ = compute_spectral_reflectivity(
            spectra, compute_power=compute_power,
            subtract_noise=subtract_noise, smooth_window=smooth_window,
            pwr_field=pwr_h_field, signal_field=signal_h_field,
            noise_field=noise_h_field)
        sdBZ_lin = np.ma.power(10., 0.1*sdBZ['data'])
        dBZ = 10.*np.ma.log10(np.ma.sum(sdBZ_lin, axis=-1))

        if ('reflectivity' in fields_list or
                'unfiltered_reflectivity' in fields_list):
            dBZ_field = 'reflectivity'
            if 'unfiltered_reflectivity' in fields_list:
                dBZ_field = 'unfiltered_'+dBZ_field
            dBZ_dict = get_metadata(dBZ_field)
            dBZ_dict['data'] = dBZ
            fields.update({dBZ_field: dBZ_dict})

    if ('reflectivity_vv' in fields_list or
            'differential_reflectivity' in fields_list or
            'velocity_vv' in fields_list or
            'spectrum_width_vv' in fields_list or
            'unfiltered_reflectivity_vv' in fields_list or
            'unfiltered_differential_reflectivity' in fields_list or
            'unfiltered_velocity_vv' in fields_list or
            'unfiltered_spectrum_width_vv' in fields_list):
        sdBZv = compute_spectral_reflectivity(
            spectra, compute_power=compute_power,
            subtract_noise=subtract_noise, smooth_window=smooth_window,
            pwr_field=pwr_v_field, signal_field=signal_v_field,
            noise_field=noise_v_field)
        sdBZv_lin = np.ma.power(10., 0.1*sdBZv['data'])
        dBZv = 10.*np.ma.log10(np.ma.sum(sdBZv_lin, axis=-1))
        if ('reflectivity_vv' in fields_list or
                'unfiltered_reflectivity_vv' in fields_list):
            dBZv_field = 'reflectivity_vv'
            if 'unfiltered_reflectivity_vv' in fields_list:
                dBZv_field = 'unfiltered_'+dBZv_field
            dBZv_dict = get_metadata(dBZv_field)
            dBZv_dict['data'] = dBZv
            fields.update({dBZv_field: dBZv_dict})

        if ('differential_reflectivity' in fields_list or
                'unfiltered_differential_reflectivity' in fields_list):
            ZDR = dBZ-dBZv

            ZDR_field = 'differential_reflectivity'
            if 'unfiltered_differential_reflectivity' in fields_list:
                ZDR_field = 'unfiltered_'+ZDR_field
            ZDR_dict = get_metadata(ZDR_field)
            ZDR_dict['data'] = ZDR
            fields.update({ZDR_field: ZDR_dict})

    if ('uncorrected_differential_phase' in fields_list or
            'uncorrected_unfiltered_differential_phase' in fields_list):
        sPhiDP = compute_spectral_differential_phase(
            spectra, use_rhohv=use_rhohv, srhohv_field=srhohv_field,
            signal_h_field=signal_h_field, signal_v_field=signal_v_field)
        sPhiDP['data'][sPhiDP['data'] < 0.] += 360.
        PhiDP = (np.ma.sum(sdBZ_lin*sPhiDP['data'], axis=-1) /
                 np.ma.sum(sdBZ_lin, axis=-1))
        PhiDP[PhiDP > 180.] -= 360.

        PhiDP_field = 'uncorrected_differential_phase'
        if 'uncorrected_unfiltered_differential_phase' in fields_list:
            PhiDP_field = 'uncorrected_unfiltered_differential_phase'
        PhiDP_dict = get_metadata(PhiDP_field)
        PhiDP_dict['data'] = PhiDP
        fields.update({PhiDP_field: PhiDP_dict})

    if ('cross_correlation_ratio' in fields_list or
            'unfiltered_cross_correlation_ratio' in fields_list):
        RhoHV_dict = compute_rhohv(
            spectra, use_rhohv=use_rhohv, subtract_noise=subtract_noise,
            srhohv_field=srhohv_field, pwr_h_field=pwr_h_field,
            pwr_v_field=pwr_v_field, signal_h_field=signal_h_field,
            signal_v_field=signal_v_field, noise_h_field=noise_h_field,
            noise_v_field=noise_v_field)
        RhoHV_field = 'cross_correlation_ratio'
        if 'unfiltered_cross_correlation_ratio' in fields_list:
            RhoHV_field = 'unfiltered_'+RhoHV_field
        fields.update({RhoHV_field: RhoHV_dict})

    if ('velocity' in fields_list or 'spectrum_width' in fields_list or
            'velocity_vv' in fields_list or
            'spectrum_width_vv' in fields_list or
            'unfiltered_velocity' in fields_list or
            'unfiltered_spectrum_width' in fields_list or
            'unfiltered_velocity_vv' in fields_list or
            'unfiltered_spectrum_width_vv' in fields_list):
        vel = np.ma.expand_dims(spectra.Doppler_velocity['data'], axis=1)
        vel = ma_broadcast_to(
            vel, (spectra.nrays, spectra.ngates, spectra.npulses_max))
        if ('velocity' in fields_list or
                'unfiltered_velocity' in fields_list or
                'spectrum_width' in fields_list or
                'unfiltered_spectrum_width' in fields_list):
            mean_vel = (
                np.ma.sum(sdBZ_lin*vel, axis=-1)/np.ma.sum(sdBZ_lin, axis=-1))
        if ('velocity_vv' in fields_list or
                'unfiltered_velocity_vv' in fields_list or
                'spectrum_width_vv' in fields_list or
                'unfiltered_spectrum_width_vv' in fields_list):
            mean_vel_v = (
                np.ma.sum(sdBZv_lin*vel, axis=-1) /
                np.ma.sum(sdBZv_lin, axis=-1))

        if 'velocity' in fields_list or 'unfiltered_velocity' in fields_list:
            vel_field = 'velocity'
            if 'unfiltered_velocity' in fields_list:
                vel_field = 'unfiltered_'+vel_field
            vel_dict = get_metadata(vel_field)
            vel_dict['data'] = mean_vel
            fields.update({vel_field: vel_dict})

        if ('velocity_vv' in fields_list or
                'unfiltered_velocity_vv' in fields_list):
            vel_field = 'velocity_vv'
            if 'unfiltered_velocity_vv' in fields_list:
                vel_field = 'unfiltered_'+vel_field
            vel_dict = get_metadata(vel_field)
            vel_dict['data'] = mean_vel_v
            fields.update({vel_field: vel_dict})

        if ('spectrum_width' in fields_list or
                'unfiltered_spectrum_width' in fields_list):
            mean_vel2 = np.ma.expand_dims(mean_vel, axis=2)
            mean_vel2 = ma_broadcast_to(
                mean_vel2,
                (spectra.nrays, spectra.ngates, spectra.npulses_max))
            width = np.ma.sqrt(
                np.ma.sum(np.ma.power(vel-mean_vel2, 2.)*sdBZ_lin, axis=-1) /
                dBZ)
            width_field = 'spectrum_width'
            if 'unfiltered_spectrum_width' in fields_list:
                width_field = 'unfiltered_'+width_field
            width_dict = get_metadata(width_field)
            width_dict['data'] = width
            fields.update({width_field: width_dict})

        if ('spectrum_width_vv' in fields_list or
                'unfiltered_spectrum_width_vv' in fields_list):
            mean_vel2 = np.ma.expand_dims(mean_vel_v, axis=2)
            mean_vel2 = ma_broadcast_to(
                mean_vel2,
                (spectra.nrays, spectra.ngates, spectra.npulses_max))
            width = np.ma.sqrt(
                np.ma.sum(np.ma.power(vel-mean_vel2, 2.)*sdBZv_lin, axis=-1) /
                dBZv)
            width_field = 'spectrum_width_vv'
            if 'unfiltered_spectrum_width_vv' in fields_list:
                width_field = 'unfiltered_'+width_field
            width_dict = get_metadata(width_field)
            width_dict['data'] = width
            fields.update({width_field: width_dict})

    radar = radar_from_spectra(spectra)

    for field_name in fields_list:
        radar.add_field(field_name, fields[field_name])

    return radar


def compute_noise_power(spectra, units='dBADU', navg=1, rmin=0.,
                        nnoise_min=1, signal_field=None):
    """
    Computes the noise power from the complex spectra in ADU.
    Requires key dBADU_to_dBm_hh or dBADU_to_dBm_vv in radar_calibration if
    the units are to be dBm. The noise is computed using the method described
    in Hildebrand and Sehkon, 1974.

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    units : str
        The units of the returned signal. Can be 'ADU', 'dBADU' or 'dBm'
    navg : int
        Number of spectra averaged
    rmin : int
        Range from which the data is used to estimate the noise
    nnoise_min : int
        Minimum number of samples to consider the estimated noise power valid
    signal_field : str, optional
        Name of the field in radar which contains the signal.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    noise_dict : field dictionary
        Field dictionary containing the noise power

    References
    ----------
    P. H. Hildebrand and R. S. Sekhon, Objective Determination of the Noise
    Level in Doppler Spectra. Journal of Applied Meteorology, 1974, 13,
    808-811.

    """
    if signal_field is None:
        signal_field = get_field_name('complex_spectra_hh_ADU')

    pol = 'hh'
    if 'vv' in signal_field:
        pol = 'vv'

    ind_rmin = np.where(spectra.range['data'] >= rmin)[0]
    if ind_rmin.size == 0:
        warn('Unable to compute spectral noise. ' +
             'Range at which start gathering data '+str(rmin) +
             'km. larger than radar range')
        return None

    ind_rmin = ind_rmin[0]

    pwr = _compute_power(spectra.fields[signal_field]['data'])

    noise = np.ma.masked_all(
        (spectra.nrays, spectra.ngates))
    for ray, npuls in enumerate(spectra.npulses['data']):
        mean, _, _, _ = estimate_noise_hs74(
            pwr[ray, ind_rmin:, 0:npuls].compressed(), navg=navg,
            nnoise_min=nnoise_min)
        noise[ray, :] = mean*npuls

    if units in ('dBADU', 'dBm'):
        noise = 10.*np.ma.log10(noise)

        if units == 'dBm':
            dBADU2dBm = None
            if spectra.radar_calibration is not None:
                if (pol == 'hh' and
                        'dBADU_to_dBm_hh' in spectra.radar_calibration):
                    dBADU2dBm = (
                        spectra.radar_calibration['dBADU_to_dBm_hh']['data'][
                            0])
                elif (pol == 'vv' and
                      'dBADU_to_dBm_vv' in spectra.radar_calibration):
                    dBADU2dBm = (
                        spectra.radar_calibration['dBADU_to_dBm_vv']['data'][0])

            if dBADU2dBm is None:
                raise ValueError(
                    'Unable to compute spectral power in dBm. ' +
                    'dBADU to dBm conversion factor unknown')

            # should it be divided by the number of pulses?
            noise += dBADU2dBm

    noise_field = 'noise'+units+'_'+pol
    noise_dict = get_metadata(noise_field)
    noise_dict['data'] = noise

    return noise_dict


def compute_reflectivity(spectra, sdBZ_field=None):
    """
    Computes the reflectivity from the spectral reflectivity

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    sdBZ_field : str
        Name of the field that contains the spectral reflectivity.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    dBZ_dict : field dictionary
        Field dictionary containing the reflectivity

    """
    if sdBZ_field is None:
        sdBZ_field = get_field_name('spectral_reflectivity_hh')

    sdBZ_lin = np.ma.power(10., 0.1*spectra.fields[sdBZ_field]['data'])
    dBZ = 10.*np.ma.log10(np.ma.sum(sdBZ_lin, axis=-1))

    dBZ_field = 'reflectivity'
    if 'vv' in sdBZ_field:
        dBZ_field += '_vv'
    if 'unfiltered' in sdBZ_field:
        dBZ_field = 'unfiltered_'+dBZ_field

    dBZ_dict = get_metadata(dBZ_field)
    dBZ_dict['data'] = dBZ

    return dBZ_dict


def compute_differential_reflectivity(spectra, sdBZ_field=None,
                                      sdBZv_field=None):
    """
    Computes the differential reflectivity from the horizontal and vertical
    spectral reflectivity

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    sdBZ_field, sdBZv_field : str
        Name of the fields that contain the spectral reflectivity.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    ZDR_dict : field dictionary
        Field dictionary containing the differential reflectivity

    """
    if sdBZ_field is None:
        sdBZ_field = get_field_name('spectral_reflectivity_hh')
    if sdBZv_field is None:
        sdBZv_field = get_field_name('spectral_reflectivity_vv')

    sdBZ_lin = np.ma.power(10., 0.1*spectra.fields[sdBZ_field]['data'])
    dBZ = 10.*np.ma.log10(np.ma.sum(sdBZ_lin, axis=-1))

    sdBZv_lin = np.ma.power(10., 0.1*spectra.fields[sdBZv_field]['data'])
    dBZv = 10.*np.ma.log10(np.ma.sum(sdBZv_lin, axis=-1))

    zdr_field = 'differential_reflectivity'
    if 'unfiltered' in sdBZ_field:
        zdr_field = 'unfiltered_'+zdr_field

    ZDR_dict = get_metadata(zdr_field)
    ZDR_dict['data'] = dBZ-dBZv

    return ZDR_dict


def compute_differential_phase(spectra, sdBZ_field=None, sPhiDP_field=None):
    """
    Computes the differential phase from the spectral differential phase and
    the spectral reflectivity

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    sdBZ_field, sPhiDP_field : str
        Name of the fields that contain the spectral reflectivity and the
        spectral differential phase.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    PhiDP_dict : field dictionary
        Field dictionary containing the differential phase

    """
    if sdBZ_field is None:
        sdBZ_field = get_field_name('spectral_reflectivity_hh')
    if sPhiDP_field is None:
        sPhiDP_field = get_field_name('spectral_differential_phase')

    sdBZ_lin = np.ma.power(10., 0.1*spectra.fields[sdBZ_field]['data'])
    sPhiDP = deepcopy(spectra.fields[sPhiDP_field]['data'])
    sPhiDP[sPhiDP < 0.] += 360.
    PhiDP = np.ma.sum(sdBZ_lin*sPhiDP, axis=-1)/np.ma.sum(sdBZ_lin, axis=-1)
    PhiDP[PhiDP > 180.] -= 360.

    phidp_field = 'uncorrected_differential_phase'
    if 'unfiltered' in sPhiDP_field:
        phidp_field = 'uncorrected_unfiltered_differential_phase'
    PhiDP_dict = get_metadata(phidp_field)
    PhiDP_dict['data'] = PhiDP

    return PhiDP_dict


def compute_rhohv(spectra, use_rhohv=False, subtract_noise=False,
                  srhohv_field=None, pwr_h_field=None, pwr_v_field=None,
                  signal_h_field=None, signal_v_field=None,
                  noise_h_field=None, noise_v_field=None):
    """
    Computes RhoHV from the horizontal and vertical spectral reflectivity
    or from sRhoHV and the spectral powers

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    use_rhohv : Bool
        If true the RhoHV will be computed from sRho_hv. Otherwise it will be
        computed using the complex spectra
    subtract_noise : Bool
            If True noise will be subtracted from the signals
    srhohv_field, pwr_h_field, pwr_v_field, signal_h_field, signal_v_field,
    noise_h_field, noise_v_field : str
        Name of the fields in radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    RhoHV_dict : field dictionary
        Field dictionary containing the RhoHV

    """
    if use_rhohv:
        if srhohv_field is None:
            srhohv_field = get_field_name(
                'spectral_copolar_correlation_coefficient')
        if pwr_h_field is None:
            pwr_h_field = get_field_name('spectral_power_hh_ADU')
        if pwr_v_field is None:
            pwr_v_field = get_field_name('spectral_power_vv_ADU')
    else:
        if signal_h_field is None:
            signal_h_field = get_field_name('complex_spectra_hh_ADU')
        if signal_v_field is None:
            signal_v_field = get_field_name('complex_spectra_vv_ADU')
        if noise_h_field is None:
            noise_h_field = get_field_name('spectral_noise_power_hh_ADU')
        if noise_v_field is None:
            noise_v_field = get_field_name('spectral_noise_power_vv_ADU')

    if not use_rhohv:
        sRhoHV = (
            spectra.fields[signal_h_field]['data'] *
            np.ma.conjugate(spectra.fields[signal_v_field]['data']))

        noise = None
        if noise_h_field in spectra.fields:
            noise = spectra.fields[noise_h_field]['data']

        pwr_h = _compute_power(
            spectra.fields[signal_h_field]['data'], noise=noise,
            subtract_noise=subtract_noise)

        noise = None
        if noise_v_field in spectra.fields:
            noise = spectra.fields[noise_v_field]['data']

        pwr_v = _compute_power(
            spectra.fields[signal_v_field]['data'], noise=noise,
            subtract_noise=subtract_noise)
    else:
        pwr_h = spectra.fields[pwr_h_field]['data']
        pwr_v = spectra.fields[pwr_v_field]['data']
        sRhoHV = spectra.fields[srhohv_field]['data']*np.ma.sqrt(pwr_h*pwr_v)

    RhoHV = (
        np.ma.abs(np.ma.sum(sRhoHV, axis=-1)) /
        np.ma.sqrt(np.ma.sum(pwr_h, axis=-1)*np.ma.sum(pwr_v, axis=-1)))

    rhohv_field = 'cross_correlation_ratio'
    if (srhohv_field is not None and 'unfiltered' in srhohv_field or
            signal_h_field is not None and 'unfiltered' in signal_h_field):
        rhohv_field = 'unfiltered_cross_correlation_ratio'

    RhoHV_dict = get_metadata(rhohv_field)
    RhoHV_dict['data'] = RhoHV

    return RhoHV_dict


def compute_Doppler_velocity(spectra, sdBZ_field=None):
    """
    Computes the Doppler velocity from the spectral reflectivity

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    sdBZ_field : str
        Name of the field that contains the spectral reflectivity.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    vel_dict : field dictionary
        Field dictionary containing the Doppler velocity

    """
    if sdBZ_field is None:
        sdBZ_field = get_field_name('spectral_reflectivity_hh')

    sdBZ_lin = np.ma.power(10., 0.1*spectra.fields[sdBZ_field]['data'])
    vel = np.ma.expand_dims(spectra.Doppler_velocity['data'], axis=1)
    vel = ma_broadcast_to(
        vel, (spectra.nrays, spectra.ngates, spectra.npulses_max))
    mean_vel = np.ma.sum(sdBZ_lin*vel, axis=-1)/np.ma.sum(sdBZ_lin, axis=-1)

    vel_field = 'velocity'
    if 'vv' in sdBZ_field:
        vel_field += '_vv'
    if 'unfiltered' in sdBZ_field:
        vel_field = 'unfiltered_'+vel_field
    vel_dict = get_metadata(vel_field)
    vel_dict['data'] = mean_vel

    return vel_dict


def compute_Doppler_width(spectra, sdBZ_field=None):
    """
    Computes the Doppler width from the spectral reflectivity

    Parameters
    ----------
    spectra : Radar spectra object
        Object containing the required fields
    sdBZ_field : str
        Name of the field that contains the spectral reflectivity.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    width_dict : field dictionary
        Field dictionary containing the Doppler spectrum width

    """
    if sdBZ_field is None:
        sdBZ_field = get_field_name('spectral_reflectivity_hh')

    sdBZ_lin = np.ma.power(10., 0.1*spectra.fields[sdBZ_field]['data'])
    dBZ = np.ma.sum(sdBZ_lin, axis=-1)

    vel = np.ma.expand_dims(spectra.Doppler_velocity['data'], axis=1)
    vel = ma_broadcast_to(
        vel, (spectra.nrays, spectra.ngates, spectra.npulses_max))

    mean_vel = np.ma.sum(sdBZ_lin*vel, axis=-1)/dBZ
    mean_vel = np.ma.expand_dims(mean_vel, axis=2)
    mean_vel = ma_broadcast_to(
        mean_vel, (spectra.nrays, spectra.ngates, spectra.npulses_max))

    width = np.ma.sqrt(
        np.ma.sum(np.ma.power(vel-mean_vel, 2.)*sdBZ_lin, axis=-1)/dBZ)

    width_field = 'spectrum_width'
    if 'vv' in sdBZ_field:
        width_field += '_vv'
    if 'unfiltered' in sdBZ_field:
        width_field = 'unfiltered_'+width_field
    width_dict = get_metadata(width_field)
    width_dict['data'] = width

    return width_dict


def _compute_power(signal, noise=None, subtract_noise=False,
                   smooth_window=None):
    """
    Compute the signal power in linear units

    Parameters
    ----------
    signal : float array
        The complex spectra
    noise : float array
        The noise power per Doppler bin
    subtract_noise : Bool
        If True and noise not None the noise power will be subtracted from the
        signal power
    smooth_window : int or None
        Size of the moving Gaussian smoothing window. If none no smoothing
        will be applied

    Returns
    -------
    pwr : float array
        The computed power

    """
    pwr = np.ma.power(np.ma.abs(signal), 2.)

    if subtract_noise and noise is not None:
        pwr -= noise
        pwr[pwr < 0.] = np.ma.masked

    if smooth_window is not None:
        pwr = _smooth_spectral_power(pwr, wind_len=smooth_window)

    return pwr


def _smooth_spectral_power(raw_data, wind_len=5):
    """
    smoothes the spectral power using a rolling Gaussian window.

    Parameters
    ----------
    raw_data : float masked array
        The data to smooth.
    wind_len : float
        Length of the moving window

    Returns
    -------
    data_smooth : float masked array
        smoothed data

    """
    # we want an odd window
    if wind_len % 2 == 0:
        wind_len += 1
    half_wind = int((wind_len-1)/2)

    # create window
    wind = gaussian(wind_len, std=1.)
    wind /= np.sum(wind)

    # initialize smoothed data
    nrays, ngates, nDoppler = np.shape(raw_data)
    data_smooth = np.ma.masked_all((nrays, ngates, nDoppler))

    # get rolling window and mask data
    data_wind = rolling_window(raw_data, wind_len)
    data_smooth[:, :, half_wind:-half_wind] = np.ma.dot(data_wind, wind)

    return data_smooth



"""
===========================
Create a two panel RHI plot
===========================

An example which creates a two panel RHI plot of a Sigmet file.  The fields
included in the two panels are reflectivity and doppler velocity.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart
import netCDF4


# read the data and create the display object
filename = 'XSW110520113537.RAW7HHL'
radar = pyart.io.read_rsl(filename)
display = pyart.graph.RadarDisplay(radar)

# fields to plot and ranges
fields_to_plot = ['reflectivity', 'velocity']
ranges = [(-32, 64), (-17.0, 17.0)]

# plot the data
nplots = len(fields_to_plot)
plt.figure(figsize=[5 * nplots, 4])

# plot each field
for plot_num in range(nplots):
    field = fields_to_plot[plot_num]
    vmin, vmax = ranges[plot_num]

    plt.subplot(1, nplots, plot_num + 1)
    display.plot(field, 0, vmin=vmin, vmax=vmax, title_flag=False)
    display.set_limits(ylim=[0, 17])

# set the figure title and show
instrument_name = radar.metadata['instrument_name'].decode('utf-8')
time_start = netCDF4.num2date(radar.time['data'][0], radar.time['units'])
time_text = ' ' + time_start.isoformat() + 'Z '
azimuth = radar.fixed_angle['data'][0]
title = 'RHI ' + instrument_name + time_text + 'Azimuth %.2f' % (azimuth)
plt.suptitle(title)
plt.show()



"""
pyart.map.polar_to_cartesian
=========================================

Routines to project polar radar data to Cartesian coordinates

.. autosummary::
    :toctree: generated/

    polar_to_cartesian
    get_earth_radius
"""
from copy import deepcopy

import numpy as np
from scipy import spatial

KE = 4 / 3.  # Constant in the 4/3 earth radius model
# Two extreme earth radius
R_EARTH_MAX = 6378.1370 * 1000
R_EARTH_MIN = 6356.7523 * 1000


def polar_to_cartesian(radar_sweep, field_name, cart_res=75,
                       max_range=None, mapping=None):
    """
    Interpolates a PPI or RHI scan in polar coordinates to a regular cartesian
    grid of South-North and West-East coordinates (for PPI) or distance at
    ground and altitude coordinates (for RHI)

    Parameters
    ----------
    radar : Radar
        Radar instance as generated py pyart
    sweep : int
        Sweep number to project to cartesian coordinates.
    field_name : str
        Name of the radar field to be interpolated
    cart_res : int, optional
        Resolution (in m.) of the cartesian grid to which polar data is
        interpolated
    max_range : int, optional
        Maximal allowed range (in m.) from radar for gates to be interpolated
    mapping : dict, optional
        Dictionnary of mapping indexes (from polar to cartesian), gets returned
        by the function (see below). Can be used as input when interpolating
        sequentially several variables for the same scan, to save significant
        time

    Returns
    -------
    coords : tuple of 2 arrays
        2D coordinates of the cartesian grid
    cart_data : 2D array
        Interpolated radar measurements (on the cartesian grid)
    mapping,: dict
        Dictionnary of mapping indexes (from polar to cartesian),which contains
        the indexes mapping the polar grid to the cartesian grid as well as some
        metadata.
    """

    # Get data to be interpolated
    pol_data = radar_sweep.get_field(0, field_name)

    is_ppi = radar_sweep.sweep_mode['data'][0] == 'ppi'

    if mapping:
        # Check if mapping is usable:
        if is_ppi != mapping['is_ppi']:
            print('Input mapping does not correspond to given scan type, ignoring it')
            mapping = None
        elif mapping['dim_pol'] != pol_data.shape:
            print('Input mapping does not correspond to dimensions of given field'
                  ', ignoring it')
            mapping = None
        else:
            cart_res = mapping['res']
            max_range = mapping['max_range']

    # Get distances of radar data
    r = radar_sweep.range['data']

    if max_range is None:
        max_range = np.max(r)

    # Cut data at max_range
    pol_data_cut = deepcopy(pol_data[:, r < max_range])
    r = r[r < max_range]

    # Set masked pixels to nan
    pol_data_cut.filled(np.nan)

    # One specificity of using the kd-tree is that we need to pad the array
    # with nans at large ranges and angles smaller and larger
    pol_data_cut = np.pad(pol_data_cut, pad_width=((1, 1), (0, 1)),
                          mode='constant', constant_values=np.nan)

    # Get angles of radar data
    if is_ppi:
        theta = radar_sweep.azimuth['data']
    else:
        theta = radar_sweep.elevation['data']

    # We need to pad theta and r as well
    theta = np.hstack([np.min(theta) - 0.1, theta, np.max(theta) + 0.1])
    r = np.hstack([r, np.max(r) + 0.1])

    r_grid_p, theta_grid_p = np.meshgrid(r, theta)

    # Generate regular cartesian grid
    if is_ppi:
        x_vec = np.arange(-max_range - cart_res,
                          max_range + cart_res, cart_res)
        y_vec = np.arange(-max_range - cart_res,
                          max_range + cart_res, cart_res)
    else:
        x_vec = np.arange(min(
            [(max_range-cart_res)*np.cos(np.radians(np.max(theta))), 0]),
                          max_range+cart_res, cart_res)

        y_vec = np.arange(0, max_range + cart_res, cart_res)

    x_grid_c, y_grid_c = np.meshgrid(x_vec, y_vec)

    if is_ppi:
        theta_grid_c = np.degrees(np.arctan2(-x_grid_c, -y_grid_c) + np.pi)
        r_grid_c = (np.sqrt(x_grid_c**2 + y_grid_c**2))
    else:
        theta_grid_c = np.degrees(-(np.arctan2(x_grid_c,
                                               y_grid_c) - np.pi / 2))
        E = get_earth_radius(radar_sweep.latitude['data'])
        r_grid_c = (np.sqrt((E * KE * np.sin(np.radians(theta_grid_c)))**2 +
                            2 * E * KE * y_grid_c + y_grid_c ** 2)
                    - E * KE * np.sin(np.radians(theta_grid_c)))

    if not mapping:
        # Kd-tree construction and query
        kdtree = spatial.cKDTree(np.vstack((r_grid_p.ravel(),
                                            theta_grid_p.ravel())).T)
        _, mapping_idx = kdtree.query(np.vstack((r_grid_c.ravel(),
                                                 theta_grid_c.ravel())).T, k=1)

        mapping = {'idx': mapping_idx, 'max_range': max_range, 'res': cart_res,
                   'is_ppi': is_ppi, 'dim_pol': pol_data.shape}

    cart_data = pol_data_cut.ravel()[mapping['idx']]
    cart_data = np.reshape(cart_data, x_grid_c.shape)

    return (x_vec, y_vec), cart_data, mapping


def get_earth_radius(latitude):
    """
    Computes the earth radius for a given latitude

    Parameters
    ----------
    latitude: latitude in degrees (WGS84)

    Returns
    -------
    earth_radius : the radius of the earth at the given latitude
    """
    a = R_EARTH_MAX
    b = R_EARTH_MIN
    num = ((a ** 2 * np.cos(latitude)) ** 2 +
           (b ** 2 * np.sin(latitude)) ** 2)
    den = ((a * np.cos(latitude)) ** 2 +
           (b * np.sin(latitude)) ** 2)

    earth_radius = np.sqrt(num / den)

    return earth_radius



"""
pyart.io.cf1
============

Utilities for reading hpl files from the Halo Photonics  streamline lidar.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

.. autosummary::
    :toctree: generated/

    read_hpl

"""

import datetime
import numpy as np
import os

from ..config import FileMetadata
from ..core import Radar
from ..io.common import make_time_unit_str

LOCATION_CAMPAIGNS = {}
LOCATION_CAMPAIGNS['Meiringen2021_a'] = [8.111, 46.741, 575.4]
LOCATION_CAMPAIGNS['Meiringen2021_b'] = [8.121, 46.740, 588]           
       
def read_hpl(filename, additional_metadata=None, location = None):
    """
    Read a hpl ASCII file from the Halo Photonoics Steamline lidar.

    Parameters
    ----------
    filename : str
        Name of CF/Radial netCDF file to read data from.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    location : str
        Name of the location from which the data comes, if provided will
        assign corresponding values of lat, lon and alt. The named locations
        can be added to the global variable LOCATION_CAMPAIGNS in this script

    Returns
    -------
    radar : Radar
        Radar object.

    Notes
    -----
    This function has not been tested on "stream" Cfradial files.

    """
    
    filemetadata = FileMetadata('hpl')
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    _range = filemetadata('range')
    _time = filemetadata('time')
    fields = {}
    fields['avg_velocity'] = filemetadata('range')
    fields['avg_reflectivity'] = filemetadata('avg_reflectivity')
    fields['absolute_beta'] = filemetadata('absolute_beta')
            
    day = datetime.datetime.strptime(os.path.basename(filename).split('_')[2],
                                     '%Y%m%d')
    
    # read header
    endheader = False
    metadata = {}
    
    azimuth['data'] = []
    elevation['data'] = []
    _time['data'] = []
    fields['avg_velocity']['data'] = []
    fields['avg_reflectivity']['data'] = []
    fields['absolute_beta']['data'] = []
    
    with open(filename) as fo:
        for i,line in enumerate(fo):
            if ':' in line:
                val = line.split(':')[1].replace('\n','').replace('\t','')
                try:
                    if '.' in val:
                        val = float(val)
                    else:
                        val = int(val)
                except:
                    pass
                metadata[line.split(':')[0]] = val
            if '***' in line:
                endheader = True
                continue
            if endheader:
                lsplit = line.split()
                if len(lsplit[0]) > 4:
                    _time['data'].append(day + datetime.timedelta(
                        hours=float(lsplit[0])))
                    azimuth['data'].append(float(lsplit[1]))
                    elevation['data'].append(float(lsplit[2]))
                    
                else:
                    fields['avg_velocity']['data'].append(float(lsplit[1]))
                    fields['avg_reflectivity']['data'].append(float(lsplit[2]))
                    fields['absolute_beta']['data'].append(float(lsplit[3]))
                    
    for k in fields.keys():
        fields[k]['data'] = np.reshape(np.array(fields[k]['data']),
                                       (len(azimuth['data']), 
                                        metadata['Number of gates']))
    
    _range['data'] = (metadata['Range gate length (m)'] * 
                      np.arange(metadata['Number of gates']))
    azimuth['data'] = np.array(azimuth['data'])
    elevation['data'] = np.array(elevation['data'])
    _time['units'] = make_time_unit_str(_time['data'][0])
    tseconds = [(d - _time['data'][0]).total_seconds() for d in _time['data']]
    _time['data'] = tseconds
    
    if type(additional_metadata) == dict:
        if 'altitude' in additional_metadata.keys():
            altitude = additional_metadata['altitude']
        if 'latitude' in additional_metadata.keys():
            latitude = additional_metadata['latitude']
        if 'longitude' in additional_metadata.keys():
            longitude = additional_metadata['longitude']
    if location in LOCATION_CAMPAIGNS.keys():
        longitude['data'] = np.array([LOCATION_CAMPAIGNS[location][0]])
        latitude['data'] = np.array([LOCATION_CAMPAIGNS[location][1]])
        altitude['data'] =  np.array([LOCATION_CAMPAIGNS[location][2]])
        
    sweep_start_ray_index['data'] = np.array([0])
    sweep_end_ray_index['data'] = np.array(
        [azimuth['data'].size-1], dtype=np.int32)
    sweep_number['data'] = np.array([0])
    
    nazi_unique = np.unique(azimuth['data'])
    nele_unique = np.unique(elevation['data'])
    if nele_unique.size == 1 and nazi_unique.size == 1:
        scan_type = 'other'
        sweep_mode['data'] = np.array(['pointing'])
        fixed_angle['data'] = nele_unique
    elif nele_unique.size == 1 and nazi_unique.size > 1:
        scan_type = 'ppi'
        sweep_mode['data'] = np.array(['azimuth_surveillance'])
        fixed_angle['data'] = nele_unique
    elif nele_unique.size > 1 and nazi_unique.size == 1:
        scan_type = 'rhi'
        sweep_mode['data'] = np.array(['elevation_surveillance'])
        fixed_angle['data'] = nazi_unique
    else:
        raise ValueError(
            'Only single sweeps PPI, RHI or pointing are supported')
   
    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation)



"""
pyart.aux_io.metranet_cartesian_reader
======================================

Routines for putting METRANET Cartesian data files into grid object.
(Used by ELDES www.eldesradar.it)

.. autosummary::
    :toctree: generated/

    read_cartesian_metranet

"""

import datetime
import platform
from warnings import warn

import numpy as np

from ..config import FileMetadata
from ..io.common import _test_arguments
from ..core.grid import Grid

from .metranet_c import get_library
from .metranet_python import read_product as read_product_python
from .metranet_c import read_product as read_product_c

# check existence of METRANET library
try:
    METRANET_LIB = get_library(momentms=False, momentpm = True)
    if platform.system() == 'Linux':
        METRANET_LIB = get_library(momentms=True)
    _METRANETLIB_AVAILABLE = True
except:
    _METRANETLIB_AVAILABLE = False

METRANET_FIELD_NAMES = {
    'NWP_HZEROCL': 'iso0_height',
    'Rain_Rate': 'radar_estimated_rain_rate', # RZC, R1F, rZC, RZF, dRZC
    'Rainfall_accumulation': 'rainfall_accumulation', # AZC, aZC, nowpal
    # 'CPC': 'rainfall_accumulation', # Check product name
    # 'CPCH': 'rainfall_accumulation', # Check product name
    # 'dACC': 'rainfall_accumulation', # Check product name
    # 'dACCH': 'rainfall_accumulation', # Check product name
    'POH': 'probability_of_hail', # BZC, dBZC
    'Hail': 'probability_of_hail', # GZC, dGZC
    'MESH': 'maximum_expected_severe_hail_size', # MZC, dMZC
    'Column_Max@swp_1_to_20': 'maximum_echo', # CZC, dCZC
    'MAXECHO_HEIGHT': 'maximum_echo_height',
    'ECHOTOP_15': 'echo_top_15dBZ', # EZC, dEZC
    'ECHOTOP_20': 'echo_top_20dBZ', # EZC
    'ECHOTOP_45': 'echo_top_45dBZ', # EZC, dEZC
    'ECHOTOP_50': 'echo_top_50dBZ', # EZC
    'VIL': 'vertically_integrated_liquid', # LZC, dLZC
    'CAPPI_Zh_1.0km': 'reflectivity', # OZC
    'CAPPI_Zh_2.0km': 'reflectivity', # OZC
    'CAPPI_Zh_3.0km': 'reflectivity', # OZC
    'CAPPI_Zh_4.0km': 'reflectivity', # OZC
    'CAPPI_Zh_5.0km': 'reflectivity', # OZC
    'CAPPI_Zh_6.0km': 'reflectivity', # OZC
    'CAPPI_Zh_7.0km': 'reflectivity', # OZC
    'CAPPI_Zh_8.0km': 'reflectivity', # OZC
    'CAPPI_Zh_9.0km': 'reflectivity', # OZC
    'CAPPI_Zh_10.0km': 'reflectivity', # OZC
    'CAPPI_Zh_11.0km': 'reflectivity', # OZC
    'CAPPI_Zh_12.0km': 'reflectivity', # OZC
    'CAPPI_Zh_13.0km': 'reflectivity', # OZC
    'CAPPI_Zh_14.0km': 'reflectivity', # OZC
    'CAPPI_Zh_15.0km': 'reflectivity', # OZC
    'CAPPI_Zh_16.0km': 'reflectivity', # OZC
    'CAPPI_Zh_17.0km': 'reflectivity', # OZC
    'CAPPI_Zh_18.0km': 'reflectivity' # OZC
}


def read_cartesian_metranet(filename, additional_metadata=None, chy0=255.,
                            chx0=-160., reader='C', **kwargs):

    """
    Read a METRANET product file.

    Parameters
    ----------
    filename : str
        Name of the METRANET file to read.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    chy0, chx0 : float
        Swiss coordinates position of the south-western point in the grid
    reader : str
        The reader library to use. Can be either 'C' or 'python'

    Returns
    -------
    grid : Grid
        Grid object containing data from METRANET file.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    if reader == 'C' and _METRANETLIB_AVAILABLE:
        ret = read_product_c(filename, physic_value=True, masked_array=True)
    elif reader == 'python':
        ret = read_product_python(
            filename, physic_value=True, masked_array=True)
    else:
        warn('Invalid reader name or C library not available,' +
             ' using python (default) instead')
        ret = read_product_python(
            filename, physic_value=True, masked_array=True)

    if ret is None:
        warn('Unable to read file '+filename)
        return None

    # reserved_variables = [
    #     'time', 'x', 'y', 'z',
    #     'origin_latitude', 'origin_longitude', 'origin_altitude',
    #     'point_x', 'point_y', 'point_z', 'projection',
    #     'point_latitude', 'point_longitude', 'point_altitude',
    #     'radar_latitude', 'radar_longitude', 'radar_altitude',
    #     'radar_name', 'radar_time', 'base_time', 'time_offset',
    #     'ProjectionCoordinateSystem']

    # metadata
    metadata = ret.header

    filemetadata = FileMetadata(
        'METRANET', METRANET_FIELD_NAMES, additional_metadata)

    nx = int(ret.header['column'])
    ny = int(ret.header['row'])
    nz = 1

    # required reserved variables
    time = filemetadata('grid_time')
    origin_latitude = filemetadata('origin_latitude')
    origin_longitude = filemetadata('origin_longitude')
    origin_altitude = filemetadata('origin_altitude')
    x = filemetadata('x')
    y = filemetadata('y')
    z = filemetadata('z')

    x['data'] = 1000.*(
        np.arange(nx)*float(ret.header['rect_xres'])+chy0 +
        float(ret.header['rect_xres'])/2.-600.)

    y['data'] = 1000.*(
        np.arange(ny)*float(ret.header['rect_yres'])+chx0 +
        float(ret.header['rect_yres'])/2.-200.)

    if ret.header['product'].startswith('CAPPI_Zh_'):
        alt = ret.header['product'].split('_')[2]
        alt = float(alt.replace('km', ''))*1000.
        z['data'] = np.array([alt])
    else:
        z['data'] = np.array([0.])

    # Origin of LV03 Swiss coordinates
    origin_latitude['data'] = np.array([46.9524055556])
    origin_longitude['data'] = np.array([7.43958333333])
    origin_altitude['data'] = np.array([0.])

    prod_time = datetime.datetime.strptime(
        ret.header['time'][0:9], '%y%j%H%M')
    time['units'] = 'seconds since '+prod_time.strftime('%Y-%m-%d %H:%M:%S')
    if 'usr_forecast_hour' in ret.header:
        time['data'] = np.array(
            [float(ret.header['usr_forecast_hour'])*3600.])
    else:
        time['data'] = np.array([0])

    # projection (Swiss Oblique Mercator)
    projection = {
        'proj': 'somerc',
        '_include_lon_0_lat_0': True
    }

    # read in the fields
    fields = {}
    field = filemetadata.get_field_name(ret.header['product'])
    field_dict = filemetadata(field)
    data = np.broadcast_to(ret.data[::-1, :], (nz, ny, nx))
    mask = np.broadcast_to(ret.data.mask[::-1, :], (nz, ny, nx))
    field_dict['data'] = np.ma.array(data, mask=mask)
    fields[field] = field_dict

    # radar variables
    radar_latitude = None
    radar_longitude = None
    radar_altitude = None
    radar_name = None
    radar_time = None

    return Grid(
        time, fields, metadata,
        origin_latitude, origin_longitude, origin_altitude, x, y, z,
        projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)



"""
pyart.correct._fourdd_interface
===============================

Cython wrapper around the University of Washington FourDD algorithm.

.. autosummary::
    :toctree: generated/

    create_soundvolume
    fourdd_dealias

"""

cimport _fourdd_h
cimport numpy as np
from pyart.io._rsl_interface cimport _RslVolume

from ..io import _rsl_interface


cpdef create_soundvolume(radialVelVolume,
                         np.ndarray[np.float32_t, ndim=1] hc,
                         np.ndarray[np.float32_t, ndim=1] sc,
                         np.ndarray[np.float32_t, ndim=1] dc,
                         maxshear=0.05, sign=1):
    """
    Create a RSL Volume containing sounding data.

    Parameters
    ----------
    radialVelVolume : _RslVolume
        Radial velocities which will be dealiased, shape used to create
        soundvolume.
    hc : ndarray
        Sounding heights in meters.  Must be a contiguous one-dimensional
        float32 array.
    sc : ndarray
        Sounding wind speed in m/s.  Must be a contiguous one-dimensional
        float32 array.
    dc : ndarray
        Sounding wind direction in degrees.  Must be a contiguous
        one-dimensional float32 array.
    maxshear : float
        Maximum vertical shear which will be incorperated into the created
        volume.
    sign : int
        Sign convention which the radial velocities in the created volume
        will follow.  A value of 1 represents when positive values
        velocities are towards the radar, -1 represents when negative
        velocities are towards the radar.

    Returns
    -------
    usuccess : int
        Flag indicating if loading of data was successful, 1 = yes, 0 = no.
    soundvolume : _RslVolume
        RslVolume containing sounding data.

    """

    cdef _RslVolume soundVolume
    cdef unsigned short success = 0
    cdef float MISSINGVEL = 131072.0
    soundVolume = _rsl_interface.copy_volume(radialVelVolume)
    success = _fourdd_h.sounding_to_volume(
        soundVolume._Volume, MISSINGVEL,
        <float *> hc.data, <float *> sc.data, <float *> dc.data,
        <int> len(hc), maxshear, sign)
    return success, soundVolume


cpdef fourdd_dealias(
    _RslVolume radialVelVolume, _RslVolume lastVelVolume,
    _RslVolume soundVolume, filt,
    compthresh=0.25, compthresh2=0.49,
    thresh=0.4, ckval=1.0, stdthresh=0.8, epsilon=0.00001, maxcount=10,
    pass2=1, rm=0, proximity=5, mingood=5, ba_mincount=5, ba_edgecount=3,
    debug=False):
    """
    fourdd_dealias(
        radialVelVolume, lastVelVolume, soundVolume, filt,
        compthresh=0.25, compthresh2=0.49, thresh=0.4,
        epsilon=0.00001, ckval=1.0, stdthresh=0.8, maxcount=10, pass2=1,
        rm=0, proximity=5, mingood=5, ba_mincount=5, ba_edgecount=3,
        debug=False)

    Dealias using the FourDD algorithm.

    Parameters
    ----------
    radialVelVolume : _RslVolume
        Radial velocities which will be dealiased.
    lastVelVolume : _RslVolume or None
        Radial velocities from a previously dealiased radar volume. For best
        results, this radar should represent the previous volume scan in time.
        If the last velocity volume is unavailable, set this to None.
    soundVolume : _RslVolume or None
        Volume created from sounding data.  If unavailable, set this to None.
        soundVolume and lastVelVolume cannot both be None.
    filt : int
        Flag controlling Bergen and Albers filter, 1 = yes, 0 = no.

    Other Parameters
    ----------------
    compthresh : float
        Fraction of the Nyquist velocity to use as a threshold when performing
        continity (initial) dealiasing.  Velocities differences above this
        threshold will not be marked as gate from which to begin unfolding
        during spatial dealiasing.
    compthresh2 : float
        The same as compthresh but the value used during the second pass of
        dealasing.  This second pass is only performed in both a sounding
        and last volume are provided.
    thresh : float
        Fraction of the Nyquist velocity to use as a threshold when performing
        spatial dealiasing.  Horizontally adjacent gates with velocities above
        this theshold will count against assigning the gate in question the
        velocity value being tested.
    ckval : float
        When the absolute value of the velocities are below this value they
        will not be marked as gates from which to begin unfolding during
        spatial dealiasing.
    stdthresh : float
       Fraction of the Nyquist velocity to use as a standard deviation
       threshold in the window dealiasing portion of the algorithm.
    epsilon : float
        Difference used when comparing a value to missing value, changing this
        from the default is not recommended.
    maxcount : int
        Maximum allowed number of fold allowed when unfolding velocities.
    pass2 : int
        Controls weather unfolded gates should be removed (a value of 0)
        or retained for unfolding during the second pass (a value of 1) when
        both a sounding volume and last volume are provided.
    rm : int
        Determines what should be done with gates that are left unfolded
        after the first pass of dealiasing.  A value of 1 will remove these
        gates, a value of 0 sets these gates to their initial velocity.  If
        both a sounding volume and last volume are provided this parameter is
        ignored.
    proximity : int
        Number of gates and rays to include of either side of the current gate
        during window dealiasing.  This value may be doubled in cases where
        a standard sized window does not capture a sufficient number of
        good valued gates.
    mingood : int
        Number of good valued gates required within the window before the
        current gate will be unfolded.
    ba_mincount : int
        Number of neighbors required during Bergen and Albers filter for
        a given gate to be included, must be between 1 and 8, 5 recommended.
    ba_edgecount : int
        Same as ba_mincount but used at ray edges, must be between 1 and 5,
        3 recommended.
    debug : bool
        True to return RSL Volume objects for debugging:
        usuccess, radialVelVolume, lastVelVolume, soundVolume, unfoldedVolume

    Returns
    -------
    usuccess : int
        Flag indicating if the unfolding was successful, 1 = yes, 0 = no.
    data : np.ndarray
        Array of unfolded velocities.

    References
    ----------
    C. N. James and R. A Houze Jr, A Real-Time Four-Dimensional Doppler
    Dealising Scheme, Journal of Atmospheric and Oceanic Technology, 2001, 18,
    1674.

    """
    cdef _RslVolume unfoldedVolume
    cdef float MISSINGVEL = 131072.0

    if lastVelVolume is None and soundVolume is None:
        raise ValueError('lastVelVolume or soundVolume must be defined')

    # The following closely follows that in FourDD.c starting at line 142

    # copy the radial velocity data to unfoldedVolume
    unfoldedVolume = _rsl_interface.copy_volume(radialVelVolume)

    # unfold the velocity fields in unfoldedVolume
    if lastVelVolume is None:   # only soundVolume
        usuccess  = _fourdd_h.dealias_fourdd(
            unfoldedVolume._Volume, soundVolume._Volume, NULL,
            MISSINGVEL, compthresh, compthresh2, thresh,
            ckval, stdthresh, epsilon,
            maxcount, pass2, rm, proximity, mingood,
            filt, ba_mincount, ba_edgecount)
    elif soundVolume is None:   # only lastVelVolume
        usuccess = _fourdd_h.dealias_fourdd(
            unfoldedVolume._Volume, NULL, lastVelVolume._Volume,
            MISSINGVEL, compthresh, compthresh2, thresh,
            ckval, stdthresh, epsilon,
            maxcount, pass2, rm, proximity, mingood,
            filt, ba_mincount, ba_edgecount)
    else:   # both soundVolume and lastVelVolume
        usuccess = _fourdd_h.dealias_fourdd(
            unfoldedVolume._Volume, soundVolume._Volume, lastVelVolume._Volume,
            MISSINGVEL, compthresh, compthresh2, thresh,
            ckval, stdthresh, epsilon,
            maxcount, pass2, rm, proximity, mingood,
            filt, ba_mincount, ba_edgecount)
    if debug:
        return (usuccess, radialVelVolume, lastVelVolume, soundVolume,
                unfoldedVolume)

    data = unfoldedVolume.get_data()
    return usuccess, data



"""
pyart.io.rsl
============

Python wrapper around the RSL library.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    _RslVolumeDataExtractor

.. autosummary::
    :toctree: generated/

    read_rsl
    VOLUMENUM2RSLNAME
    RSLNAME2VOLUMENUM

"""

import datetime
import warnings

import numpy as np

from ..config import FileMetadata, get_fillvalue
try:
    from . import _rsl_interface
    _RSL_AVAILABLE = True
except ImportError:
    _RSL_AVAILABLE = False
from ..core.radar import Radar
from .common import make_time_unit_str
from ..lazydict import LazyLoadDict
from ..exceptions import MissingOptionalDependency


def read_rsl(filename, field_names=None, additional_metadata=None,
             file_field_names=False, exclude_fields=None,
             delay_field_loading=False, include_fields=None,
             radar_format=None, callid=None, skip_range_check=False):
    """
    Read a file supported by RSL.

    Parameters
    ----------
    filename : str or RSL_radar
        Name of file whose format is supported by RSL.
    field_names : dict, optional
        Dictionary mapping RSL data type names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the RSL data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed. In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects.
    radar_format : str or None
        Format of the radar file. Must be 'wsr88d' or None.
    callid : str or None
        Four letter NEXRAD radar Call ID, only used when radar_format is
        'wsr88d'.
    skip_range_check : bool, optional
        True to skip check for uniform range bin location, the reported range
        locations will only be verified true for the first ray. False will
        perform the check and raise a IOError when the locations of the gates
        change between rays.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # check that TRMM RSL is available
    if not _RSL_AVAILABLE:
        raise MissingOptionalDependency(
            "Py-ART must be build with support for TRMM RSL to use the " +
            "read_rsl function.")

    # create metadata retrieval object
    filemetadata = FileMetadata('rsl', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # read the file, determine common parameters
    fillvalue = get_fillvalue()
    rslfile = _rsl_interface.RslFile(
        filename.encode('ascii'), radar_format, callid)
    available_vols = rslfile.available_moments()
    first_volume = rslfile.get_volume(available_vols[0])
    first_sweep = first_volume.get_sweep(0)
    first_ray = first_sweep.get_ray(0)
    nsweeps = first_volume.nsweeps

    # scan_type, sweep_mode, fixed_angle
    fixed_angle = filemetadata('fixed_angle')
    fdata = first_volume.get_sweep_fix_angles()
    fdata[fdata < 0.] = 360. + fdata[fdata < 0.]    # all angles positive
    fixed_angle['data'] = fdata
    sweep_mode = filemetadata('sweep_mode')

    if rslfile.scan_mode == 0:      # 0 = PPI, 1 = RHI
        scan_type = 'ppi'
        sweep_mode['data'] = np.array(
            nsweeps * ['azimuth_surveillance'], dtype='S')
    else:
        scan_type = 'rhi'
        sweep_mode['data'] = np.array(nsweeps * ['rhi'], dtype='S')

    # time
    time = filemetadata('time')
    t_start = datetime.datetime(
        rslfile.year, rslfile.month, rslfile.day, rslfile.hour,
        rslfile.minute, int(rslfile.sec), int(1e6 * (rslfile.sec % 1)))
    datetimes = []
    for i in range(nsweeps):
        sweep = first_volume.get_sweep(i)
        for j in range(sweep.nrays):
            datetimes.append(sweep.get_ray(j).get_datetime())
    t_delta = [t-t_start for t in datetimes]
    sec_since_start = [
        t.seconds + t.days*3600*24 + t.microseconds/1.e6 for t in t_delta]
    time['data'] = np.array(sec_since_start, dtype=np.float64)
    time['units'] = make_time_unit_str(t_start)

    # range
    if (not skip_range_check) and (not first_volume.is_range_bins_uniform()):
        message = (
            "Range bin locations change between rays. File cannot be read "
            "with with correct range locations for all rays. "
            "To read in data reporting the ranges from the first ray set the "
            "**skip_range_check** parameter to True.")
        raise IOError(message)

    _range = filemetadata('range')
    gate0 = first_ray.range_bin1
    gate_size = first_ray.gate_size
    ngates = first_ray.nbins
    _range['data'] = gate0 + gate_size * np.arange(ngates, dtype='float32')
    _range['meters_to_center_of_first_gate'] = _range['data'][0]
    _range['meters_between_gates'] = np.array(gate_size, dtype='float32')

    # fields
    # transfer only those which are available
    fields = {}
    for volume_num in available_vols:
        # if a volume number is not recognized it is skipped.
        if volume_num not in VOLUMENUM2RSLNAME:
            warnings.warn("Unknown Volume Number %d" % volume_num)
            continue
        rsl_field_name = VOLUMENUM2RSLNAME[volume_num]
        field_name = filemetadata.get_field_name(rsl_field_name)
        if field_name is None:
            continue

        # create the field dictionary
        field_dic = filemetadata(field_name)
        field_dic['_FillValue'] = fillvalue
        data_extractor = _RslVolumeDataExtractor(
            rslfile, volume_num, fillvalue)
        if delay_field_loading:
            field_dic = LazyLoadDict(field_dic)
            field_dic.set_lazy('data', data_extractor)
        else:
            field_dic['data'] = data_extractor()
        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata('metadata')
    metadata['original_container'] = 'rsl'
    rsl_dict = rslfile.get_radar_header()
    need_from_rsl_header = {
        'name': 'instrument_name', 'project': 'project', 'state': 'state',
        'country': 'country'}  # rsl_name : radar_metadata_name
    for rsl_key, metadata_key in need_from_rsl_header.items():
        metadata[metadata_key] = rsl_dict[rsl_key]

    # latitude
    latitude = filemetadata('latitude')
    lat = _dms_to_d((rsl_dict['latd'], rsl_dict['latm'], rsl_dict['lats']))
    latitude['data'] = np.array([lat], dtype='float64')

    # longitude
    longitude = filemetadata('longitude')
    lon = _dms_to_d((rsl_dict['lond'], rsl_dict['lonm'], rsl_dict['lons']))
    longitude['data'] = np.array([lon], dtype='float64')

    # altitude
    altitude = filemetadata('altitude')
    altitude['data'] = np.array([rsl_dict['height']], dtype='float64')

    # sweep_number, sweep_mode, sweep_start_ray_index, sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    sweep_number['data'] = np.arange(nsweeps, dtype='int32')
    ray_count = first_volume.get_nray_array()   # array of rays in each sweep
    ssri = np.cumsum(np.append([0], ray_count[:-1])).astype('int32')
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = np.cumsum(ray_count).astype('int32') - 1

    # azimuth, elevation
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')
    _azimuth, _elevation = first_volume.get_azimuth_and_elev_array()
    azimuth['data'] = _azimuth
    elevation['data'] = _elevation

    # instrument_parameters
    prt = filemetadata('prt')
    prt_mode = filemetadata('prt_mode')
    nyquist_velocity = filemetadata('nyquist_velocity')
    unambiguous_range = filemetadata('unambiguous_range')
    beam_width_h = filemetadata('radar_beam_width_h')
    beam_width_v = filemetadata('radar_beam_width_v')

    pm_data, nv_data, pr_data, ur_data = first_volume.get_instr_params()
    prt['data'] = pr_data
    prt_mode['data'] = pm_data.astype('S')
    nyquist_velocity['data'] = nv_data
    unambiguous_range['data'] = ur_data
    beam_width_h['data'] = np.array([first_sweep.horz_half_bw * 2.],
                                    dtype='float32')
    beam_width_v['data'] = np.array([first_sweep.vert_half_bw * 2.],
                                    dtype='float32')

    instrument_parameters = {'unambiguous_range': unambiguous_range,
                             'prt_mode': prt_mode, 'prt': prt,
                             'nyquist_velocity': nyquist_velocity,
                             'radar_beam_width_h': beam_width_h,
                             'radar_beam_width_v': beam_width_v}

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def _dms_to_d(dms):
    """ Degrees, minutes, seconds to degrees """
    return dms[0] + (dms[1] + dms[2] / 60.0) / 60.0


class _RslVolumeDataExtractor(object):
    """
    Class facilitating on demand extraction of data from a RSL file.

    Parameters
    ----------
    rslfile : RslFile
        Open RslFile object to extract data from.
    volume_num : int
        Volume number of data to be extracted.
    fillvalue : int
        Value used to fill masked values in the returned array.

    """

    def __init__(self, rslfile, volume_num, fillvalue):
        """ initialize the object. """
        self.rslfile = rslfile
        self.volume_num = volume_num
        self.fillvalue = fillvalue

    def __call__(self):
        """ Return an array containing data from the referenced volume. """
        # extract the field and mask
        data = self.rslfile.get_volume_array(self.volume_num)
        data[np.where(np.isnan(data))] = self.fillvalue
        data[np.where(data == 131072)] = self.fillvalue
        return np.ma.masked_equal(data, self.fillvalue)


VOLUMENUM2RSLNAME = {
    0: 'DZ',
    1: 'VR',
    2: 'SW',
    3: 'CZ',
    4: 'ZT',
    5: 'DR',
    6: 'LR',
    7: 'ZD',
    8: 'DM',
    9: 'RH',
    10: 'PH',
    11: 'XZ',
    12: 'CD',
    13: 'MZ',
    14: 'MD',
    15: 'ZE',
    16: 'VE',
    17: 'KD',
    18: 'TI',
    19: 'DX',
    20: 'CH',
    21: 'AH',
    22: 'CV',
    23: 'AV',
    24: 'SQ',
    25: 'VS',
    26: 'VL',
    27: 'VG',
    28: 'VT',
    29: 'NP',
    30: 'HC',
    31: 'VC',
    32: 'V2',
    33: 'S2',
    34: 'V3',
    35: 'S3',
    36: 'CR',
    37: 'CC',
    38: 'PR',
    39: 'SD',
    40: 'ZZ',
    41: 'RD',
    42: 'ET',
    43: 'EZ',
}

RSLNAME2VOLUMENUM = dict([(v, k) for k, v in VOLUMENUM2RSLNAME.items()])



"""
========================================
Radar Retrievals (:mod:`pyart.retrieve`)
========================================

.. currentmodule:: pyart.retrieve

Radar retrievals.

Radar retrievals
================

.. autosummary::
    :toctree: generated/

    kdp_maesaka
    kdp_schneebeli
    kdp_vulpiani
    kdp_leastsquare_single_window
    kdp_leastsquare_double_window
    calculate_snr_from_reflectivity
    calculate_velocity_texture
    get_ml_rng_limits
    get_iso0_val
    compute_ccor
    compute_snr
    compute_l
    compute_cdr
    compute_noisedBZ
    compute_radial_noise_hs
    compute_radial_noise_ivic
    compute_signal_power
    compute_rcs
    compute_rcs_from_pr
    compute_vol_refl
    compute_bird_density
    fetch_radar_time_profile
    map_profile_to_gates
    steiner_conv_strat
    hydroclass_semisupervised
    data_for_centroids
    select_samples
    compute_centroids
    determine_medoids
    synthetic_obs_distribution
    _destandardize
    get_freq_band
    texture_of_complex_phase
    grid_displacement_pc
    grid_shift
    est_rain_rate_zpoly
    est_rain_rate_z
    est_rain_rate_kdp
    est_rain_rate_a
    est_rain_rate_zkdp
    est_rain_rate_za
    est_rain_rate_hydro
    est_wind_vel
    est_vertical_windshear
    atmospheric_gas_att
    get_coeff_attg
    est_wind_profile
    detect_ml
    melting_layer_giangrande
    melting_layer_hydroclass
    melting_layer_mf
    compute_apparent_profile
    _get_res_vol_sides
    velocity_azimuth_display
    quasi_vertical_profile
    compute_qvp
    compute_rqvp
    compute_evp
    compute_svp
    compute_vp
    compute_ts_along_coord
    compute_iq
    compute_spectral_power
    compute_spectral_noise
    compute_spectral_phase
    compute_spectral_reflectivity
    compute_spectral_differential_reflectivity
    compute_spectral_differential_phase
    compute_spectral_rhohv
    compute_pol_variables
    compute_noise_power
    compute_reflectivity
    compute_differential_reflectivity
    compute_differential_phase
    compute_rhohv
    compute_Doppler_velocity
    compute_Doppler_width
    compute_reflectivity_iq
    compute_st1_iq
    compute_st2_iq
    compute_wbn_iq
    compute_differential_reflectivity_iq
    compute_mean_phase_iq
    compute_differential_phase_iq
    compute_rhohv_iq
    compute_Doppler_velocity_iq
    compute_Doppler_width_iq
    compute_pol_variables_iq
    compute_spectra
    gecsx

"""
from .ml import detect_ml, melting_layer_giangrande, melting_layer_hydroclass
from .ml import _get_res_vol_sides, compute_apparent_profile, melting_layer_mf
from .ml import get_ml_rng_limits, get_iso0_val
from .kdp_proc import kdp_maesaka, kdp_schneebeli, kdp_vulpiani
from .kdp_proc import kdp_leastsquare_single_window
from .kdp_proc import kdp_leastsquare_double_window
from .echo_class import steiner_conv_strat, hydroclass_semisupervised
from .echo_class import get_freq_band, data_for_centroids, compute_centroids
from .echo_class import select_samples, determine_medoids, _destandardize
from .echo_class import synthetic_obs_distribution
from .gate_id import map_profile_to_gates, fetch_radar_time_profile
from .simple_moment_calculations import calculate_snr_from_reflectivity
from .simple_moment_calculations import calculate_velocity_texture
from .simple_moment_calculations import compute_snr, compute_l, compute_cdr
from .simple_moment_calculations import compute_noisedBZ, compute_signal_power
from .simple_moment_calculations import get_coeff_attg, compute_vol_refl
from .simple_moment_calculations import compute_bird_density
from .simple_moment_calculations import atmospheric_gas_att, compute_ccor
from .simple_moment_calculations import compute_rcs, compute_rcs_from_pr
from .simple_moment_calculations import compute_radial_noise_hs
from .simple_moment_calculations import compute_radial_noise_ivic
from .qpe import est_rain_rate_z, est_rain_rate_zpoly, est_rain_rate_kdp
from .qpe import est_rain_rate_a, est_rain_rate_zkdp, est_rain_rate_za
from .qpe import est_rain_rate_hydro
from .advection import grid_displacement_pc, grid_shift
from .wind import est_wind_vel, est_vertical_windshear, est_wind_profile
from .vad import velocity_azimuth_display
from .qvp import quasi_vertical_profile, compute_qvp, compute_rqvp
from .qvp import compute_evp, compute_svp, compute_vp, compute_ts_along_coord
from .spectra import compute_spectral_power, compute_spectral_phase
from .spectra import compute_spectral_noise, compute_spectral_reflectivity
from .spectra import compute_spectral_differential_reflectivity
from .spectra import compute_spectral_differential_phase
from .spectra import compute_spectral_rhohv, compute_reflectivity
from .spectra import compute_differential_reflectivity
from .spectra import compute_differential_phase, compute_rhohv
from .spectra import compute_Doppler_velocity, compute_Doppler_width
from .spectra import compute_pol_variables, compute_iq, compute_noise_power
from .iq import compute_reflectivity_iq, compute_differential_reflectivity_iq
from .iq import compute_differential_phase_iq, compute_rhohv_iq
from .iq import compute_Doppler_velocity_iq, compute_Doppler_width_iq
from .iq import compute_pol_variables_iq, compute_spectra
from .iq import compute_mean_phase_iq, compute_st1_iq, compute_st2_iq
from .iq import compute_wbn_iq
from .gecsx import gecsx

__all__ = [s for s in dir() if not s.startswith('_')]



"""
==================================
Map two radars to a Cartesian grid
==================================

Map the reflectivity field of two nearby ARM XSARP radars from antenna
coordinates to a Cartesian grid.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart

# read in the data from both XSAPR radars
XSAPR_SW_FILE = 'swx_20120520_0641.nc'
XSAPR_SE_FILE = 'sex_20120520_0641.nc'
radar_sw = pyart.io.read_cfradial(XSAPR_SW_FILE)
radar_se = pyart.io.read_cfradial(XSAPR_SE_FILE)

# filter out gates with reflectivity > 100 from both radars
gatefilter_se = pyart.filters.GateFilter(radar_se)
gatefilter_se.exclude_transition()
gatefilter_se.exclude_above('corrected_reflectivity_horizontal', 100)
gatefilter_sw = pyart.filters.GateFilter(radar_sw)
gatefilter_sw.exclude_transition()
gatefilter_sw.exclude_above('corrected_reflectivity_horizontal', 100)

# perform Cartesian mapping, limit to the reflectivity field.
grid = pyart.map.grid_from_radars(
    (radar_se, radar_sw), gatefilters=(gatefilter_se, gatefilter_sw),
    grid_shape=(1, 201, 201),
    grid_limits=((1000, 1000), (-50000, 40000), (-60000, 40000)),
    grid_origin = (36.57861, -97.363611),
    fields=['corrected_reflectivity_horizontal'])

# create the plot
fig = plt.figure()
ax = fig.add_subplot(111)
ax.imshow(grid.fields['corrected_reflectivity_horizontal']['data'][0],
          origin='lower', extent=(-60, 40, -50, 40), vmin=0, vmax=48)
plt.show()



""" Unit Tests for Py-ART's io/arm_sonde.py module. """

import datetime

import numpy as np
from numpy.testing import assert_almost_equal
from pytest import raises

import pyart


def test_read_arm_sonde_vap_target_datetime():
    target_datetime = datetime.datetime(2011, 5, 10, 11, 30, 5)
    profile_datetime, hprofile = pyart.io.read_arm_sonde_vap(
        pyart.testing.INTERP_SOUNDE_FILE, target_datetime=target_datetime)

    assert profile_datetime == datetime.datetime(2011, 5, 10, 11, 30)
    assert_almost_equal(hprofile.height[:5], [318, 338, 358, 378, 398], 1)
    assert_almost_equal(hprofile.speed[:5], [8.4, 7.4, 9.2, 10.8, 12.3], 1)
    assert_almost_equal(
        hprofile.direction[:5], [182.6, 181.9, 182.3, 183.3, 184.2], 1)


def test_read_arm_sonde_vap_radar():
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    radar.time['units'] = 'seconds since 2011-05-10T11:30:05Z'
    profile_datetime, hprofile = pyart.io.read_arm_sonde_vap(
        pyart.testing.INTERP_SOUNDE_FILE, radar=radar)

    assert profile_datetime == datetime.datetime(2011, 5, 10, 11, 30)
    assert_almost_equal(hprofile.height[:5], [318, 338, 358, 378, 398], 1)
    assert_almost_equal(hprofile.speed[:5], [8.4, 7.4, 9.2, 10.8, 12.3], 1)
    assert_almost_equal(
        hprofile.direction[:5], [182.6, 181.9, 182.3, 183.3, 184.2], 1)


def test_read_arm_sonde_vap_errors():

    # radar or target_datetime must be specified
    raises(
        ValueError, pyart.io.read_arm_sonde_vap,
        pyart.testing.INTERP_SOUNDE_FILE)

    # only one of radar or target_datetime can be specified
    radar = pyart.testing.make_empty_ppi_radar(1, 1, 1)
    target_datetime = datetime.datetime(2011, 5, 10, 11, 30, 5)
    raises(
        ValueError, pyart.io.read_arm_sonde_vap,
        pyart.testing.INTERP_SOUNDE_FILE,
        radar=radar, target_datetime=target_datetime)


def test_read_arm_sonde():
    profile_dt, hprofile = pyart.io.read_arm_sonde(pyart.testing.SONDE_FILE)

    assert profile_dt == datetime.datetime(2011, 5, 20, 8, 28)
    assert_almost_equal(hprofile.height[:5], [315, 321, 328, 336, 344], 0)
    assert_almost_equal(hprofile.speed[:5], [5, 3.2, 3.7, 4.3, 4.8], 1)
    assert_almost_equal(
        hprofile.direction[:5], [215., 193., 191., 191., 189.], 1)



""" Unit tests for Py-ART's util/simulated_vel.py module. """

import numpy as np
from numpy.testing import assert_almost_equal
import pytest

import pyart

@pytest.mark.skip(
    reason=("Test does not fail locally, but does when "
            "pytest is ran on all of Py-ART. Can't find issue."))
def test_simulated_vel():

    radar = pyart.testing.make_target_radar()

    # profile of 10 m/s winds out of the west at all heights
    height = np.arange(0, 1000, 30)
    speed = np.ones_like(height) * 10.
    direction = np.ones_like(height) * 270.
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)

    sim_vel = pyart.util.simulated_vel_from_profile(radar, profile)

    assert sim_vel['data'].shape == (360, 50)

    # check simulated velocities along the north, east, south and west rays
    assert_almost_equal(sim_vel['data'][[0, 90, 180, 270], 10],
                        [0, 10, 0, -10], 0)



"""
pyart.retrieve._gecsx_functions
==================

Helping functions for visibility and ground echoes estimation from a DEM.

"""

import warnings
import logging
import numpy as np
from scipy.special import erfc


from ._gecsx_functions_cython import vis_weighting

RCS_MIN = 1.0 # [m^2] remove RCS below this level
def antenna_pattern_gauss(d_az, d_el, antenna_3dB, db = False,
    twoway = True, az_conv = None, el_conv = None, units = 'rad'):
    """
    Get the antenna weighting factor due to the azimuth and elevation offsets
    from the main antenna direction. The weighting factor is meant in terms of
    power (not amplitude of the E-field). An Gaussian antenna pattern is
    assumed.

    Parameters
    ----------
    d_az : array
        Azimuth offsets to the main antenna axis [deg]
    d_el : array
        Elevation offset to the main antenna axis [deg]
    antenna_3dB : float
        Half power beam width [deg]
    db : bool (optional)
        If true return the result in dB instead linear.
    twoway: bool (optional)
        If true, return the two-way weighting factor instead of
        the one-way factor.
    az_conv (optional): float
        If set, assumes that the antenna moves in azimuth direction (PPI) and
        averages over the angle given by this keyword [deg].
    el_conv (optional): float
        If set, assumes that the antenna moves in elevation direction (RHI)
        and averages over the angle given by this keyword [deg].
    units (optional) : str
        Specify if inputs quantities are given in radians ( "rad" ) or
        degrees ( "deg" )

    Returns
    -------
    fa : array
        Weighting factor.
    """
    if az_conv != None:
        if az_conv <= 0:
            az_conv = None
    if el_conv != None:
        if el_conv <= 0:
            el_conv = None

    if units not in ['deg','rad']:
        print('Invalid units, must be either "rad" or "deg", '+
                                              'assuming they are "rad"')

    if units == 'deg':
        # Convert all quantities to rad
        if az_conv is not None:
            az_conv *= np.pi / 180.
        if el_conv is not None:
            el_conv *= np.pi / 180.
        d_az = d_az.copy() * np.pi / 180.
        d_el = d_el.copy() * np.pi / 180.
        antenna_3dB *= np.pi / 180.

    if az_conv == None or el_conv == None:
        if az_conv is not  None:
            """ The antenna is moving in azimuth direction and is averaging the
             received pulses over 'az_conv' deg. The norm azimuth position
             of the antenna is reached, when the antenna moved half of the
             azimuth distance (az_offset = 'az_conv'/2 deg).
             The weighting factor at the azimuth position 'daz' from the norm
             position is given by the following integral:

                                1    / daz+az_offset
               fa(daz, del) = ---- * |               f(daz) d(daz)   * f(del)
                              Norm   / daz-az_offset
             where
               daz : Is the azimuth deviation from the norm antenna position
               del : Is the elevation deviation from the norm antenna position
               fa(daz,del) : Weighting factor at point (daz,del) f(0,0) must be 1.
               Norm : Normalization such that f(0,0)=1
               f(x) : Weighting factor of the non moving antenna (Gaussian
                     function, see below)

             Solving the integral above leads to:

                             K1
             fa(daz, del) = ---- * ( erf(K*(daz+az_offset)) -erf(K*(daz-az_offset)) )  * f(del)
                            Norm
             where
                   2 * sqrt(ln(2))
               K = ---------------
                       phi3db

                    sqrt(!PI)
               K1 = ---------
                      2 * K

               erf : the error function
               phi3db : the half power beam width
            """

            az_offset = az_conv/2.
            K = 2. * np.sqrt(np.log(2)) / antenna_3dB
            K1 = np.sqrt(np.pi) / 2. / K
            Norm = 2.0 * K1 * erfc(K * az_offset)
            faz = (K1 / Norm*(erfc(K * (d_az + az_offset)) -
                              erfc(K * (d_az - az_offset))))
        else:
            da = (2. * d_az / antenna_3dB) ** 2
            ind = da > 20.
            da[ind] = 20
            faz = np.exp(-da * np.log10(2))

        if el_conv is not None:
            # see explanation for az_conv above
            el_offset = el_conv/2. * np.pi / 180.
            K = 2. * np.sqrt(np.log(2)) / antenna_3dB
            K1 = np.sqrt(np.pi) / 2. / K
            Norm = 2.0 * K1 * erfc(K * el_offset)
            fel = (K1 / Norm*(erfc(K * (d_el + el_offset)) -
                                erfc(K * (d_el - el_offset))))
        else:
            de = (2. * d_el / antenna_3dB) ** 2
            ind = de > 20.
            de[ind] = 20
            fel = np.exp(-de * np.log10(2))

        fa = faz * fel
    else:
        # Gaussian antenna pattern:
        #
        # f(daz,del) = e^(-( (2*daz/phi3db_el)^2 + (2*del/phi3db_az)^2 ) * ln(2))
        #
        # from Gauss normal distribution N(x) with N(x)=1 and N(x=X0/2)=1/2 :
        #
        # N(x) = e^(-(2*x/X0)^2 * ln(2))

        da = 2. * d_az / antenna_3dB
        de = 2. * d_el / antenna_3dB
        dr = (da ** 2 + de ** 2)

        ind = dr > 20.
        dr[ind] = 20

        fa = np.exp(-dr * np.log(2))

    if twoway:
        fa = fa ** 2

    if db:
        fa = 10.* np.log10(fa)

    return fa


def clip_grid(grid, xr, yr, extra_m = 5000):
    """
    Clips a grid by limiting to an area defined by vectors xr and yr

    Parameters
    ----------
    grid : pyart.Core.Grid
        grid object that contains all Cartesian fields to be clipped
    xr : array
        1D array of x-coordinates outside of which the grid will be clipped
    extra_m : int
        Additional distance outside of domain defined by xr and yr which will
        be kept in the grid, i.e. the lower left corner of the final grid will
        be (min(xr) - extra_m, min(yr) - extra_m) and the upper right corner
        will be (max(xr) + extra_m, max(yr) + extra_m)

    Returns
    -------
    grid : pyart.Core.Grid
        The clipped grid
    """

    min_x = np.min(xr)
    min_y = np.min(yr)
    max_x = np.max(xr)
    max_y = np.max(yr)

    mask_x = np.logical_and(grid.x['data']  > min_x - extra_m,
                            grid.x['data']  < max_x + extra_m)
    mask_y = np.logical_and(grid.y['data']  > min_y - extra_m,
                            grid.y['data']  < max_y + extra_m)

    grid.x['data'] = grid.x['data'][mask_x]
    grid.y['data'] = grid.y['data'][mask_y]
    for f in grid.fields.keys():
        nz = len(grid.fields[f]['data']) # Nb of z levels
        grid.fields[f]['data'] = grid.fields[f]['data'][np.ix_(range(nz),
                                                      mask_y, mask_x)]
    grid.nx = len(grid.x['data'])
    grid.ny = len(grid.y['data'])
    return grid


def range_weights(rangemap, rr, pulselength, db = False):
    """
    Get the contribution of a radar target at distance 'r' to the radar signal
    at distance 'R'. Assuming a rectangular pulse and a matched filter with
    the same shape as the radar pulse. Assuming that the radar receiver is not
    averaging over multiple samples around R. The contribution is linear with
    1 at the r=R and 0 if |r - R| >= pulselength.

    Parameters
    ----------
    rangemap : array
        Array with distances (r) to the radar [m]
    rr : float
        Distance of the reference point to the radar [m]
    pulselength : float
        Pulse length of the radar pulse [m]
    db : bool
        If true return the result in dB instead linear.

    Returns
    -------
    fr : array
        Weighting factors for each range point.
    """

    dr = rangemap - rr
    fr = 1. - np.abs(dr) / pulselength

    ind0 = fr < 0
    fr[ind0] = 0.
    fr[~ind0] = fr[~ind0] * (rr / rangemap[~ind0]) ** 4

    if db:
        fr = 10.* np.log10(fr)

    return fr

def rcs(azmap, rmap, elmap, areaeffmap, sigma0map, vismap, rpol, azpol,
        elpol, DEM_res,  DEM_xmin,  DEM_ymin,  rad_x,  rad_y,   beamwidth,
        pulsewidth,  range_weighting = True,  az_conv = 0,
        rcs_raster_oversampling = 1, verbose = True):
    """
    Computes the radar cross section of ground clutter in polar coordinates

    Parameters
    ----------
    azmap : array
        Cartesian array with radar azimuth angles
    elmap : array
        Cartesian array with radar elevation angles
    rmap : array
        Cartesian array with distances from radar
    areaeffmap : array
        Cartesian array with effective areas
    sigma0map : array
        Cartesian array with sigma0 : the ratio between effective backscatter-
        ing areas and RCS
    vismap : array
        Cartesian array with radar visibility
    rpol : array
        Radar ranges at which to compute the RCS in polar coordinates
    azpol : array
        Radar azimuths at which to compute the RCS in polar coordinates
    elpol : array
        Radar elevations at which to compute the RCS in polar coordinates
    DEM_res : int
        Resolution of the DEM grid
    DEM_xmin : float
        minimum x coord of the DEM (W-E)
    DEM_ymin : float
        minimum y coord of the DEM (S-N)
    rad_x : float
        Radar position x coordinate
    rad y : float
        Radar position y coordinate
    beamwidth : float
        Antenna 3dB beamwidth in deg
    pulsewidth : float
        Radar pulsewidth in sec
    range_weighting : bool, optional
        If true, weight the backscattering cross sections of each cell
        according to their range offset using the radar pulse length.
    az_conv : float, optional
        If set, assumes that the antenna moves in azimuth direction
        (PPI) and averages over the angle given by this keyword [deg].
    rcs_raster_oversampling : int (optional)
        The raster resolution of the DEM should be smaller than
        the range resolution of the radar (defined by the pulse length).
        If this is not the case, this keyword can be set to increase the
        raster resolution. The values for the elevation, sigma naught,
        visibility are repeated. The other values are recalculated.
        Values for RCS_RASTER_OVERSAMPLING:
        0 or undefined: No oversampling is done
        1: Oversampling is done. The factor N is automatically calculated
        such that 2*dx/N < pulse length
        2 or larger: Oversampling is done with this value as N
    verbose : bool, optional
        If true, will print info about current progress

    Returns
    -------
    rcspolarmap : array
        2D Array of RCS in polar coordinates in PyART format, with all az
        angles for all elevations stacked
    """

    nrows, ncols = azmap.shape
    area_unweighted = areaeffmap * sigma0map

    pulselength = pulsewidth * 3.e8 /2. # [m]
    if az_conv is not None:
        az_conv_offset = az_conv/2.
    else:
        az_conv_offset = 0

    beamwidth_rad = beamwidth * np.pi / 180.

    if not range_weighting:
        range_weight = 1 # unity

    if rcs_raster_oversampling == 0:
        N = 1
    elif rcs_raster_oversampling == 1:
        N = int(np.ceil(2 * DEM_res / pulselength))
    else:
        N = rcs_raster_oversampling

    if N != 1:
        # New dimensions
        nc = N * ncols
        nr = N * nrows

        # repeat the values NxN, equivalent of rebin in IDL
        elvals   = np.repeat(np.repeat(elmap, N, axis=0), N, axis=1)
        areavals = np.repeat(np.repeat(area_unweighted / N ** 2,
                                       N, axis=0), N, axis=1)
        visvals  = np.repeat(np.repeat(vismap, N, axis=0), N, axis=1)

        # New x- and y-vectors
        xvec =  np.arange(nc) * DEM_res / N + DEM_xmin
        yvec =  np.arange(nr) * DEM_res / N + DEM_ymin

        xdiff = (xvec - rad_x)
        ydiff = (yvec - rad_y)

        # New distance from radar map
        X, Y = np.meshgrid(xdiff, ydiff)
        rvals  = np.sqrt( X ** 2 + Y ** 2)

        # New azimuth map
        azmap_rad =  (np.arctan2(X, Y) + 2*np.pi) % (2*np.pi)
        azvals = azmap_rad * 180. / np.pi
    else:
        rvals    = rmap
        azvals = azmap
        azmap_rad = azvals * np.pi / 180.
        elvals   = elmap
        areavals = area_unweighted
        visvals  = vismap

    elmap_rad = elvals * np.pi / 180.
    elevations_rad = np.array(elpol) * np.pi / 180.

    # Define the area around a point P(range, azimuth) where the cells
    # have a contribution to the RCS. This area is defined with the
    # range limits from range-dr_offset to range+dr_offset and the
    # azimuth limits from azimuth-daz_offset to azimuth+daz_offset.
    #
    # For a Gaussian antenna, azimuth offset more than 2*HPBW does not a
    # have remarkable contribution.
    # With a rectangular pulse and a matched filter cells farer away
    # than pulse length does not a have remarkable contribution.

    daz_offset = (2. * beamwidth)  + az_conv_offset # [deg]
    dr_offset  = pulselength # [m]

    nazim = len(azpol)
    nrange = len(rpol)
    # pyart storage format: 2D arrays (naz * nel, nranges)
    rcspolarmap = np.zeros((nazim * len(elpol), nrange)) + np.nan

    for rind in range(nrange):
        if verbose:
            logging.info('Computing range bin {:2.1f}'.format(rpol[rind]))
        rr = rpol[rind]

        indr =  np.logical_and(np.logical_and(rvals >= rr - dr_offset,
                            rvals < rr + dr_offset),
                            visvals > 0)

        if not np.any(indr):
            continue

        indr = np.where(indr)

        for azind in range(nazim):
            az = azpol[azind]
            # Inside the loops over range (rr) and azimuth (az), the
            # coordinates (rr, az) describe the point P(rr, az) for which
            # the RCS is calculated. If more than one DEM cell is within
            # the area from az-daz/2 to az+daz/2 and from rr-dr/2 to
            # rr+dr/2, the calculated RCS value is set to all of these
            # cells (next neighbor).

            # Get area around rr and az
            azmin = az - daz_offset
            azmax = az + daz_offset
            if azmin < 0:
                azmin = 360. + azmin
                indaz = np.logical_or(np.logical_and(azvals[indr] >= 0,
                                              azvals[indr] < azmax),
                          np.logical_and(azvals[indr] >= azmin,
                                             azvals[indr] <= 360.))
            elif azmax > 360:
                azmax = azmax - 360.
                indaz = np.logical_or(np.logical_and(azvals[indr] >= azmin,
                                             azvals[indr] <= 360),
                                  np.logical_and(azvals[indr] >= 0,
                                             azvals[indr] < azmax))
            else:
                indaz = np.logical_and(azvals[indr] >= azmin,
                                       azvals[indr] < azmax)

            inda = tuple([indr[0][indaz],
                          indr[1][indaz]])    # Cells that contribute to the cells to set indset

            # Calculate offsets in azimuth and elevation to the
            # point P(rr,az) and the elevation angle of the antenna.

            daz_area = azmap_rad[inda] - (az * np.pi / 180.)

            indaz = daz_area > np.pi
            daz_area[indaz] = daz_area[indaz] - 2. * np.pi

            indaz = daz_area < -np.pi
            daz_area[indaz] = daz_area[indaz] + 2. * np.pi

            if range_weighting:
                # Get the weighting factor due to the range offset.
                range_weight = range_weights(rvals[inda], rr,
                                                pulselength)

            ind_rzero = rvals[inda] <= 0.0
            if np.any(ind_rzero) :
                continue

            for iel, el in enumerate(elevations_rad):
                del_area = elmap_rad[inda] - el


                # Get the two-way weighting factor due to the azimuth offset
                # to the main antenna direction (assuming a Gaussian antenna pattern).
                ant_weight = antenna_pattern_gauss(daz_area, del_area,
                                            beamwidth_rad, twoway = True,
                                            az_conv = az_conv * np.pi / 180.,
                                            units = 'rad')

                # RCS = SUM_j sigma_j
                # = SUM_j sigma0_j * A_eff_j * fa(dphi_j,dteta_j)^2 * fr(drange)
                # where
                # sigma_j  : Backscattering cross section of each cell [m^2]
                # sigma0_j : Sigma naught of each cell [1]
                # A_eff_j  : Effective area of each cell [m^2]
                # fa       : One-way weighting function due to the azimuth and elevation offsets.
                # fr       : Range weighting function due to the range offset

                # RCS contribution of each cell inside the contribution
                # area.
                rcs_area = ant_weight * range_weight * areavals[inda]
                # Sum up all the contributions
                rcs = np.nansum(rcs_area)


                if rcs < RCS_MIN:
                    rcs = np.nan

                # Set rcs to all values inside the set area.

                rcspolarmap[azind + iel * nazim, rind] = rcs
    return rcspolarmap


def sigma0(inc_ang, frequency, method = 'Gabella'):
    """
    Estimates the sigma0 factor (ratio between effective backscattering area)
    and Radar Cross Section

    Parameters
    ----------
    inc_ang : array
        Incidence angles [deg]
    frequency_ghz : float
        Frequency in GHz
    method : str, optional
        Which estimation method to use, either 'Gabella' or 'Delrieu'

    Returns
    -------
    sigma_0 : array
        RCS to Backscattering area ratio

    References
    ----------
    Delrieu, G., J. D. Creutin, and H. Andrieu, 1995: Simulation of radar
    mountain returns using a digitized terrain model.
    J. Atmos. Oceanic Technol.,12, 1038–1049.

    Gabella, M., & Perona, G. (1998). Simulation of the Orographic Influence
    on Weather Radar Using a Geometric–Optics Approach, Journal of Atmospheric
    and Oceanic Technology, 15(6), 1485-1494.
    """

    inc_angr = inc_ang * np.pi / 180.
    sigma_0 = np.zeros((inc_ang.shape))

    if method not in ['Gabella', 'Delrieu']:
        warnings.warn('Invalid method for sigma_0: use "gabella" or '+\
                      '"delrieu", using "Gabella" instead')
        method = 'Gabella'

    if method == 'Gabella':
        # GABELLA/PERONA 1997 (GECS)
        te1 = 80.     # Maximum Angle of "Plateau" Region      [deg]
        te2 = 87.     # Dimming Start Angle at Grazing Angles  [deg]
        te3 = 88.     # Border Incidence Angle                 [deg]
        k1 = 0.1      # Offset Index
        k2 = 1.       # Dimming Index near Grazing Angles
        bbdeg = 2.75  # Tuning Angle in Exp Rising above Border Angle [deg]

        te1r = te1 * np.pi / 180.0
        te2r = te2 * np.pi / 180.0
        te3r = te3 * np.pi / 180.0
        bbdegr = bbdeg * np.pi / 180.0

        factor =  k1 * np.cos(te2r) * ((np.pi / 2. - te2r) /
                                       (np.pi / 2.-te1r) ) ** k2

        ind0 = inc_ang <= te1

        if len(ind0) != 0:
            sigma_0[ind0] = k1 * np.cos(inc_angr[ind0])

        ind1 = np.logical_and(inc_ang > te1, inc_ang <= te2)

        if len(ind1) != 0:
            iang = inc_angr[ind1]
            sigma_0[ind1] =  k1 * np.cos(iang) * ((np.pi / 2. - iang) /
                                           (np.pi / 2.-te1r) ) ** k2

        ind2 = np.logical_and(inc_ang > te2, inc_ang <= te3)

        if len(ind2) != 0:
            sigma_0[ind2] = factor

        ind3 = np.logical_and(inc_ang > te3, inc_ang <= 90)

        if len(ind3) != 0:
            iang = inc_angr[ind3]
            sigma_0[ind3] = factor * np.exp((iang - te3r) / bbdegr)

        ind4 = inc_ang > 90

        if len(ind4) != 0:
            sigma_0[ind4] = factor * np.exp((np.pi / 2. - te3r)
                                                   / bbdegr)

    elif method == 'Delrieu':
        # DELRIEU 1995 (X-BAND) / SKOLNIK 1990
        lim_ang_del = 63.75  # Borderline Angle separating curve branche
        # First Branch  [0..lim_ang_del]
        a1 = -9.1    # Skolnik's Model "a1" parameter [dB]
        b1 = -0.12   # Skolnik's Model "b1" parameter [dB/deg]
        c1 = 0.25    # Skolnik's Model "c1" parameter [dB/GHz]
        d1 = 0.      # Skolnik's Model "d1" parameter [dB/(deg*GHz))]
        # Second Branch  [lim_ang_del..90]
        a1_2 = 12.93 # Skolnik's Model "a1" parameter [dB]
        b1_2 = -0.37 # Skolnik's Model "b1" parameter [dB/deg]
        c1_2 = 0.    # Skolnik's Model "c1" parameter [dB/GHz]
        d1_2 = 0.    # Skolnik's Model "d1" parameter [dB/(deg*GHz))]

        sigma_db = np.zeros(sigma_0.shape) + np.nan

        ind = inc_ang < lim_ang_del

        if len(ind) != 0:
            iang = inc_angr[ind]
            sigma_db[ind] = (a1 + b1 * iang + c1 * frequency +
                             d1 * frequency * iang)
        ind =inc_ang >= lim_ang_del

        if len(ind) != 0:
            iang = inc_angr[ind]
            sigma_db[ind] = (a1_2 + b1_2 * iang + c1_2 * frequency +
                             d1_2 * frequency * iang)

        sigma_0 = 10 ** (sigma_db / 10.)

    return sigma_0


def visibility(azmap, rmap, elmap, DEM_res, DEM_xmin, DEM_ymin, rad_x, rad_y,
               dr, daz, verbose = True):
    """
    Computes the radar visibility over the DEM grid (in Cartesian coords)

    Parameters
    ----------
    azmap : array
        Cartesian array with radar azimuth angles
    rmap : array
        Cartesian array with distances from radar
    elmap : array
        Cartesian array with radar elevation angles
    DEM_res : int
        Resolution of the DEM grid
    DEM_xmin : float
        minimum x coord of the DEM (W-E)
    DEM_ymin : float
        minimum y coord of the DEM (S-N)
    rad_x : float
        Radar position x coordinate
    rad y : float
        Radar position y coordinate
    dr : float
        Range discretization used when computing the Cartesian visibility field
        the larger the better but the slower the processing will be
    daz : float
        Azimuth discretization used when computing the Cartesian visibility
        field, the larger the better but the slower the processing will be
    verbose : bool, optional
        If true, will print info about current progress

    Returns
    -------
    visibility : array
        2D Array of visibility along the Cartesian DEM grid, 1 for pixels
        which are visible, zero otherwise.
    minviselev : array
        2D array of minimum visible elevation angle along the Cartesian
        DEM grid.
    """

    range_max = np.max(rmap)
    nr = int(range_max / dr)
    nrows = len(elmap)
    ncols = len(elmap[0])
    visib = np.zeros([nrows, ncols])
    minviselev = np.zeros([nrows, ncols])

    rr_start = dr / 2

    azmin = 0.
    azmax = 0.
    rrmin = 0.
    rrmax = 0.
    azmin_sin = 0.
    azmax_sin = 0.
    azmin_cos = 0.
    azmax_cos = 0.
    el_rmin_azmin = 0.
    el_rmin_azmax = 0.
    el_rmax_azmin = 0.
    el_rmax_azmax = 0.
    el_max = 0.
    el_max_prev = 0.

    kx_rmin_azmin = 0
    kx_rmin_azmax = 0
    kx_rmax_azmin = 0
    kx_rmax_azmax = 0
    ky_rmin_azmin = 0
    ky_rmin_azmax = 0
    ky_rmax_azmin = 0
    ky_rmax_azmax = 0

    radkx = int(np.round((rad_x - DEM_xmin) / DEM_res))
    radky = int(np.round((rad_y - DEM_ymin) / DEM_res))
    for kx in range(radkx-1, radkx + 2):
        for ky in range(radky-1, radky + 2):
            visib[kx,ky] = 1
            minviselev[kx,ky] =  elmap[kx, ky]

    az_ = np.arange(0,360 + daz, daz)

    for azind in range(len(az_)):
        if verbose:
            logging.info('Computing azimuth {:2.1f}'.format(az_[azind]))
        az = az_[azind]
        azmin = az - daz/2.
        azmax = az + daz/2.
        if azmin < 0:
            azmin = 360. + azmin
            indseta = np.logical_or(np.logical_and(azmap >= 0,
                                                  azmap < azmax),
                              np.logical_and(azmap >= azmin,
                                             azmap <= 360.))
        elif azmax > 360:
            azmax = azmax - 360.
            indseta = np.logical_or(np.logical_and(azmap >= azmin,
                                                   azmap <= 360),
                              np.logical_and(azmap >= 0,
                                             azmap < azmax))
        else:
            indseta = np.logical_and(azmap >= azmin,
                                     azmap < azmax)

        azmin_sin = np.sin(azmin * np.pi / 180.)
        azmax_sin = np.sin(azmax * np.pi / 180.)
        azmin_cos = np.cos(azmin * np.pi / 180.)
        azmax_cos = np.cos(azmax * np.pi / 180.)

        indseta = np.where(indseta)

        el_max_prev = -90
        for rind in range(nr):
            rr = rr_start + rind * dr

            rrmin = rr - dr/2
            rrmax = rr + dr/2

            if np.any(indseta):

                indsetr = np.logical_and(rmap[indseta] >= rrmin,
                                         rmap[indseta] < rrmax)

                indsetr = tuple([indseta[0][indsetr],
                            indseta[1][indsetr]]) # Cells to set

                kx_rmin_azmin = radkx + int(round((rrmin * azmin_sin)/DEM_res))
                kx_rmin_azmax = radkx + int(round((rrmin * azmax_sin)/DEM_res))
                kx_rmax_azmin = radkx + int(round((rrmax * azmin_sin)/DEM_res))
                kx_rmax_azmax = radkx + int(round((rrmax * azmax_sin)/DEM_res))

                ky_rmin_azmin = radky + int(round((rrmin * azmin_cos)/DEM_res))
                ky_rmin_azmax = radky + int(round((rrmin * azmax_cos)/DEM_res))
                ky_rmax_azmin = radky + int(round((rrmax * azmin_cos)/DEM_res))
                ky_rmax_azmax = radky + int(round((rrmax * azmax_cos)/DEM_res))

                el_rmin_azmin = -90.
                el_rmin_azmax = -90.
                el_rmax_azmin = -90.
                el_rmax_azmax = -90.

                if ((kx_rmin_azmin >= 0) and (kx_rmin_azmin < ncols) and
                   (ky_rmin_azmin >= 0) and (ky_rmin_azmin < nrows)):
                    el_rmin_azmin = elmap[ky_rmin_azmin, kx_rmin_azmin]
                if ((kx_rmin_azmax >= 0) and (kx_rmin_azmax < ncols) and
                   (ky_rmin_azmax >= 0) and (ky_rmin_azmax < nrows)):
                    el_rmin_azmin = elmap[ky_rmin_azmax, kx_rmin_azmax]
                if ((kx_rmax_azmin >= 0) and (kx_rmax_azmin < ncols) and
                   (ky_rmax_azmin >= 0) and (ky_rmax_azmin < nrows)):
                    el_rmin_azmin = elmap[ky_rmax_azmin, kx_rmax_azmin]
                if ((kx_rmax_azmax >= 0) and (kx_rmax_azmax < ncols) and
                   (ky_rmax_azmax >= 0) and (ky_rmax_azmax < nrows)):
                    el_rmin_azmin = elmap[ky_rmax_azmax, kx_rmax_azmax]

                el_max = max([el_rmin_azmin, el_rmin_azmax,
                             el_rmax_azmin, el_rmax_azmax])

                if np.any(indsetr):
                    el_max = max([el_max, np.max(elmap[indsetr])])
                    if el_max >= el_max_prev:
                        visib[indsetr] = 1
                        minviselev[indsetr] = el_max
                    else:
                        minviselev[indsetr] = el_max_prev

                el_max = max([el_max, el_max_prev])
                el_max_prev = el_max

    return visib, minviselev



def visibility_angle(minviselmap, azmap, rmap,
                     rpol, azpol,  elpol, DEM_res,  DEM_xmin,  DEM_ymin,
                     rad_x, rad_y, beamwidth, pulsewidth, az_conv = 0,
                     raster_oversampling = 1, verbose = True):
    """
    Computes the radar visibility in polar coordinates

    Parameters
    ----------
    minviselmap : array
        Cartesian array with minimum visible elevation angles
    azmap: array
        Cartesian array with azimuth angles
    rmap : array
        Cartesian array with distances from radar
    rpol : array
        Radar ranges at which to compute the RCS in polar coordinates
    azpol : array
        Radar azimuths at which to compute the RCS in polar coordinates
    elpol : array
        Radar elevations at which to compute the RCS in polar coordinates
    DEM_res : int
        Resolution of the DEM grid
    DEM_xmin : float
        minimum x coord of the DEM (W-E)
    DEM_ymin : float
        minimum y coord of the DEM (S-N)
    rad_x : float
        Radar position x coordinate
    rad y : float
        Radar position y coordinate
    beamwidth : float
        Antenna 3dB beamwidth in deg
    pulsewidth : float
        Radar pulsewidth in sec
    az_conv : float, optional
        If set, assumes that the antenna moves in azimuth direction
        (PPI) and averages over the angle given by this keyword [deg].
    raster_oversampling : int (optional)
        The raster resolution of the DEM should be smaller than
        the range resolution of the radar (defined by the pulse length).
        If this is not the case, this keyword can be set to increase the
        raster resolution. The values for the elevation, sigma naught,
        visibility are repeated. The other values are recalculated.
        Values for raster_oversampling:
        0 or undefined: No oversampling is done
        1: Oversampling is done. The factor N is automatically calculated
        such that 2*dx/N < pulse length
        2 or larger: Oversampling is done with this value as N
    verbose : bool, optional
        If true, will print info about current progress

    Returns
    -------
    vispol : array
        2D Array of RCS in polar coordinates in PyART format, with all az
        angles for all elevations stacked
    """
    ncols, nrows = minviselmap.shape

    pulselength = pulsewidth * 3.e8 /2. # [m]
    az_conv_offset = az_conv/2.

    if raster_oversampling == 0:
        N = 1
    elif raster_oversampling == 1:
        N = int(np.ceil(2 * DEM_res / pulselength))
    else:
        N = raster_oversampling

    if N != 1:
        # New dimensions
        nc = N * ncols
        nr = N * nrows

        # repeat the values NxN, equivalent of rebin in IDL
        minvisvals  = np.repeat(np.repeat(minviselmap, N, axis=0), N, axis=1)

        # New x- and y-vectors
        xvec =  np.arange(nc) * DEM_res / N + DEM_xmin
        yvec =  np.arange(nr) * DEM_res / N + DEM_ymin

        xdiff = (xvec - rad_x)
        ydiff = (yvec - rad_y)

        # New distance from radar map
        X, Y = np.meshgrid(xdiff, ydiff)
        rvals  = np.sqrt( X ** 2 + Y ** 2)

        # New azimuth map
        azmap_rad =  (np.arctan2(X, Y) + 2*np.pi) % (2*np.pi)
        azvals = azmap_rad * 180. / np.pi
    else:
        rvals    = rmap
        azvals = azmap
        minvisvals   = minviselmap

    """
    Define the area around a point P(range, azimuth) where the cells
    have a contribution to the visibility. This area is defined with the
    range limits from range-dr/2 to range+dr/2 and the
    azimuth limits from azimuth-daz_offset to azimuth+daz_offset.

    For a Gaussian antenna, azimuth offset more than 2*HPBW does not a
    have remarkable contribution.
    With a rectangular pulse and a matched filter cells farer away
    than pulse length does not a have remarkable contribution.
    """

    daz_offset =  (2. * beamwidth)  + az_conv_offset # [deg]
    del_offset =  (2. * beamwidth) # [deg]

    delta_deg = 0.1 # [deg]
    ndaz = int(2 * daz_offset / delta_deg)
    ndel = int(2 * del_offset / delta_deg)

    daz_vec = np.arange(ndaz + 1) * delta_deg - daz_offset  # [deg]
    del_vec = np.arange(ndel + 1) * delta_deg - del_offset  # [deg]

    daz_area_antenna, del_area_antenna = np.meshgrid(daz_vec, del_vec)

    # Get the two-way weighting factor due to the azimuth and elevation offsets
    # to the main antenna direction (assuming a Gaussian antenna pattern).
    ant_weight = antenna_pattern_gauss(daz_area_antenna, del_area_antenna,
                                          beamwidth, az_conv = az_conv,
                                          units = 'deg')

    ant_weight_total = np.nansum(ant_weight)

    nazim = len(azpol)
    nrange = len(rpol)
    range_resolution = rpol[1] - rpol[0]

    # pyart storage format: 2D arrays (naz * nel, nranges)
    vispol = np.zeros((nazim * len(elpol), nrange))

    # Inside the loops over range (rr) and azimuth (az), the
    # coordinates (rr, az) describe the point P(rr, az) for which
    # the visibility is calculated. If more than one DEM cell is within
    # the area from az-daz/2 to az+daz/2 and from rr-dr/2 to
    # rr+dr/2, the calculated visibility value is set to all of these
    # cells (next neighbor).

    for iaz in range(nazim):
        logging.info('Computing azimuth {:2.1f}'.format(azpol[iaz]))
        # Get azimuth values to explore
        azmin = azpol[iaz] - daz_offset
        azmax = azpol[iaz] + daz_offset
        if azmin < 0:
            azmin = 360. + azmin
            indaz = np.logical_or(np.logical_and(azvals >= 0, azvals < azmax),
                    np.logical_and(azvals >= azmin, azvals <= 360.))

        elif azmax > 360:
            azmax = azmax - 360.
            indaz = np.logical_or(np.logical_and(azvals >= azmin,
                                                 azvals <= 360),
                                    np.logical_and(azvals >= 0,
                                               azvals < azmax))
        else:
            indaz = np.logical_and(azvals >= azmin, azvals < azmax)

        if not np.any(indaz):
            logging.warning('Visibility for azim {:f} not known'.format(azpol[iaz]))
            continue

        indaz = np.where(indaz)

        # These declaration avoids reindexing at every loop turn
        # and save a lot of time
        rvals_indaz = rvals[indaz]

        for iel, el in enumerate(elpol):
            # range bins (center pulse) [m]rvec[ir]-dr/2
            rmin_ground = rpol * np.cos(el * np.pi / 180.)
            # range bins (center pulse) [m]rvec[ir]+dr/2
            rmax_ground = (rpol + range_resolution) * np.cos(el * np.pi / 180.)

            for ir in range(0, nrange):
                # Get range values to explore
                indr = np.logical_and(rvals_indaz >= rmin_ground[ir],
                                      rvals_indaz <  rmax_ground[ir])


                if not np.any(indr):
                    logging.warning('Visibility for az {:f} deg and range {:f} not known'
                          .format(azpol[iaz], rpol[ir]))
                    vispol[iaz + iel * nazim, ir] = 100.
                    continue

                # DEM cells that contribute to the cells to set indset
                indcells = tuple([indaz[0][indr], indaz[1][indr]])

                ind = minvisvals[indcells] < el - del_offset
                if np.all(ind): # radar beam completely above min vis ang
                    if ir > 0:
                        if vispol[iaz + iel * nazim, ir - 1] == 100:
                            vispol[iaz + iel * nazim, ir] = 100
                        else:
                            vispol[iaz + iel * nazim,
                                   ir] = vispol[iaz + iel * nazim, ir - 1]
                    else:
                        vispol[iaz + iel * nazim, ir] = 100.
                    continue

                ind = minvisvals[indcells] > el + del_offset
                if np.all(ind): # radar beam completely below min vis ang
                    continue

                # Calculate offsets in azimuth to the point P(rr,az)
                daz_area = azvals[indcells] - azpol[iaz]

                ind = daz_area > 180.
                daz_area[ind] = daz_area[ind] - 360.
                ind = daz_area < -180.
                daz_area[ind] = daz_area[ind] + 360.

                vis = 0.
                ind_rzero = rvals[indcells] <= 0.0

                if not np.any(ind_rzero):
                    vis = vis_weighting(daz_vec, del_vec, daz_area, ant_weight,
                                        minvisvals[indcells], el)

                    vis = (vis / ant_weight_total) * 100.


                # Set vis to all values inside the set area.
                if ir > 0:
                    if vispol[iaz + iel * nazim, ir - 1] < vis:
                        vispol[iaz + iel * nazim,
                                    ir] = vispol.data[iaz + iel * nazim,
                                                           ir - 1]
                    else:
                        vispol[iaz + iel * nazim, ir] = vis
                else:
                    vispol[iaz + iel * nazim, ir] = vis

    return  vispol
